from typing import Optional, cast, List, TypeVar

import rewrite.java as j
from rewrite import Tree, list_map
from rewrite.java import J, Assignment, JLeftPadded, AssignmentOperation, MemberReference, MethodInvocation, \
    MethodDeclaration, Empty, ArrayAccess, Space, If, Block, ClassDeclaration, VariableDeclarations, JRightPadded, \
    Import, ParameterizedType, Parentheses, Try, ControlParentheses, J2
from rewrite.python import PythonVisitor, SpacesStyle, Binary, ChainedAssignment, Slice, CollectionLiteral, \
    ForLoop, DictLiteral, KeyValue, TypeHint, MultiImport, ExpressionTypeTree, ComprehensionExpression, NamedArgument
from rewrite.visitor import P, Cursor


class SpacesVisitor(PythonVisitor):
    def __init__(self, style: SpacesStyle, stop_after: Optional[Tree] = None):
        self._style = style
        self._before_parentheses = style.before_parentheses
        self._stop_after = stop_after

    def visit_class_declaration(self, class_declaration: ClassDeclaration, p: P) -> J:
        c = cast(ClassDeclaration, super().visit_class_declaration(class_declaration, p))

        # Handle space before parentheses for class declaration e.g. class A () <-> class A()
        if c.padding.implements is not None:
            c = c.padding.with_implements(
                space_before_container(c.padding.implements, self._style.before_parentheses.method_call)
            )

            param_size = len(c.padding.implements.elements)
            use_space = self._style.within.method_call_parentheses

            # TODO: Refactor to remove duplicate code.
            def _process_argument(index, arg, args_size):
                if index == 0:
                    arg = arg.with_element(space_before(arg.element, use_space))
                else:
                    arg = arg.with_element(space_before(arg.element, self._style.other.after_comma))

                if index == args_size - 1:
                    arg = space_after(arg, use_space)
                else:
                    arg = space_after(arg, self._style.other.before_comma)

                return arg

            if c.implements:
                c = c.padding.with_implements(
                    c.padding.implements.padding.with_elements(
                        list_map(lambda arg, index: _process_argument(index, arg, param_size),
                                 c.padding.implements.padding.elements)))
        return c

    def visit_method_declaration(self, method_declaration: MethodDeclaration, p: P) -> J:
        m: MethodDeclaration = cast(MethodDeclaration, super().visit_method_declaration(method_declaration, p))

        m = m.padding.with_parameters(
            space_before_container(m.padding.parameters, self._style.before_parentheses.method_declaration)
        )

        param_size = len(m.parameters)
        use_space = self._style.within.method_call_parentheses

        def _process_argument(index, arg, args_size):
            # TODO: Fix Type hint handling
            # Handle space before type hint colon e.g. foo(a: int) <-> foo(a:int)
            if isinstance(arg.element, VariableDeclarations) and arg.element.type_expression:
                vd = cast(VariableDeclarations, arg.element)
                arg = arg.with_element(vd.with_type_expression(
                    space_before(vd.type_expression, self._style.other.after_colon))
                                       .padding.with_variables(
                    list_map(lambda v: space_after_right_padded(v, self._style.other.before_colon),
                             vd.padding.variables))
                                       )

            if index == 0:
                arg = arg.with_element(space_before(arg.element, use_space))
            else:
                arg = arg.with_element(space_before(arg.element, self._style.other.after_comma))

            if index == args_size - 1:
                arg = space_after(arg, use_space)
            else:
                arg = space_after(arg, self._style.other.before_comma)
            return arg

        m = m.padding.with_parameters(
            m.padding.parameters.padding.with_elements(
                list_map(lambda arg, idx: _process_argument(idx, arg, param_size),
                         m.padding.parameters.padding.elements)
            )
        )

        if m.return_type_expression is not None:
            m = m.with_return_type_expression(space_before(m.return_type_expression, True))

        # TODO: Type parameters are not supported yet
        if m.padding.type_parameters is not None:
            raise NotImplementedError("Type parameters are not supported yet")

        return m.padding.with_parameters(
            m.padding.parameters.with_before(
                Space.SINGLE_SPACE if self._before_parentheses.method_declaration else Space.EMPTY
            )
        )

    def visit_catch(self, catch: Try.Catch, p: P) -> J:
        c = cast(Try.Catch, super().visit_catch(catch, p))
        # c = c.with_parameter(c.parameter.with_tree(space_before(c.parameter.tree, True)))
        return c

    def visit_control_parentheses(self, control_parentheses: ControlParentheses[J2], p: P) -> J:
        cp = cast(ControlParentheses[J2], super().visit_control_parentheses(control_parentheses, p))
        cp = space_before(cp, False)
        cp = cp.with_tree(space_before(cp.tree, True))
        return cp

    def visit_named_argument(self, named_argument: NamedArgument, p: P) -> J:
        a = cast(NamedArgument, super().visit_named_argument(named_argument, p))
        if a.padding.value is not None:
            a = a.padding.with_value(
                space_before_left_padded(a.padding.value, self._style.around_operators.eq_in_keyword_argument))
            return a.padding.with_value(
                space_before_left_padded_element(a.padding.value, self._style.around_operators.eq_in_keyword_argument))
        return a

    @staticmethod
    def _part_of_method_header(cursor: Cursor) -> bool:
        if (c := cursor.parent_tree_cursor()) and isinstance(c.value, VariableDeclarations):
            return c.parent_tree_cursor() is not None and isinstance(c.parent_tree_cursor().value, MethodDeclaration)
        return False

    def visit_variable(self, named_variable: VariableDeclarations.NamedVariable, p: P) -> J:
        v = cast(VariableDeclarations.NamedVariable, super().visit_variable(named_variable, p))

        # Check if the variable is a named parameter in a method declaration
        if not self._part_of_method_header(self.cursor):
            return v

        if v.padding.initializer is not None and v.padding.initializer.element is not None:
            use_space = self._style.around_operators.eq_in_named_parameter or v.variable_type is not None
            # Argument with a typehint will always receive a space e.g. foo(a: int =1) <-> foo(a: int = 1)
            use_space |= self.cursor.first_enclosing_or_throw(VariableDeclarations).type_expression is not None

            v = v.padding.with_initializer(
                space_before_left_padded(v.padding.initializer, use_space))
            v = v.padding.with_initializer(
                space_before_left_padded_element(v.padding.initializer, use_space))
        return v

    def visit_block(self, block: Block, p: P) -> J:
        b = cast(Block, super().visit_block(block, p))
        b = space_before(b, self._style.other.before_colon)
        return b

    def visit_method_invocation(self, method_invocation: MethodInvocation, p: P) -> J:
        m: MethodInvocation = cast(MethodInvocation, super().visit_method_invocation(method_invocation, p))

        # Handle space before parenthesis for method e.g. foo (..) <-> foo(..)
        m = m.padding.with_arguments(m.padding.arguments.with_before(
            Space.SINGLE_SPACE if self._style.before_parentheses.method_call else Space.EMPTY))

        if not m.arguments or isinstance(m.arguments[0], Empty):
            # Handle within method call parenthesis with no arguments e.g. foo() <-> foo( )
            use_space = self._style.within.empty_method_call_parentheses
            m = m.padding.with_arguments(
                m.padding.arguments.padding.with_elements(
                    list_map(lambda arg: arg.with_element(space_before(arg.element, use_space)),
                             m.padding.arguments.padding.elements)
                )
            )
        else:
            # Handle:
            # - Within method call parenthesis with arguments e.g. foo( 1, 2 ) <-> foo(1, 2).
            #   Note: this only applies to the space left of the first argument and right of the last argument.
            # - Space after comma e.g. foo(1,2,3) <-> foo(1, 2, 3)
            args_size = len(m.arguments)
            use_space = self._style.within.method_call_parentheses

            def _process_argument(index, arg, args_size, use_space):
                if index == 0:
                    arg = arg.with_element(space_before(arg.element, use_space))
                else:
                    arg = arg.with_element(space_before(arg.element, self._style.other.after_comma))

                if index == args_size - 1:
                    arg = space_after(arg, use_space)
                else:
                    arg = space_after(arg, self._style.other.before_comma)

                return arg

            m = m.padding.with_arguments(
                m.padding.arguments.padding.with_elements(
                    list_map(lambda arg, idx: _process_argument(idx, arg, args_size, use_space),
                             m.padding.arguments.padding.elements)))

        # TODO: Handle type parameters, relevant for constructors in Python.
        return m

    def visit_array_access(self, array_access: ArrayAccess, p: P) -> J:
        a: ArrayAccess = cast(ArrayAccess, super().visit_array_access(array_access, p))
        use_space_within_brackets = self._style.within.brackets
        index_padding = a.dimension.padding.index
        element_prefix = update_space(index_padding.element.prefix, use_space_within_brackets)
        index_after = update_space(index_padding.after, use_space_within_brackets)

        a = a.with_dimension(
            a.dimension.padding.with_index(
                index_padding.with_element(
                    index_padding.element.with_prefix(element_prefix)
                ).with_after(index_after)
            ).with_prefix(update_space(a.dimension.prefix, self._style.before_parentheses.left_bracket))
        )
        return a

    def visit_assignment(self, assignment: Assignment, p: P) -> J:
        """
        Handle assignment operator e.g. a = 1 <-> a=1
        """
        a: Assignment = cast(Assignment, super().visit_assignment(assignment, p))

        # ignore assignments of the form `<value> as x` in `with` statements
        if isinstance(self.cursor.parent_tree_cursor().value, j.Try.Resource):
            return a

        a = a.padding.with_assignment(
            space_before_left_padded(a.padding.assignment, self._style.around_operators.assignment))
        a = a.padding.with_assignment(
            a.padding.assignment.with_element(
                space_before(a.padding.assignment.element, self._style.around_operators.assignment)))
        return a

    def visit_assignment_operation(self, assignment_operation: AssignmentOperation, p: P) -> J:
        """
        Handle assignment operation e.g. a += 1 <-> a+=1
        """
        a: AssignmentOperation = cast(AssignmentOperation, super().visit_assignment_operation(assignment_operation, p))
        operator: JLeftPadded = a.padding.operator
        a = a.padding.with_operator(
            operator.with_before(update_space(operator.before, self._style.around_operators.assignment)))
        return a.with_assignment(space_before(a.assignment, self._style.around_operators.assignment))

    def visit_chained_assignment(self, chained_assignment: ChainedAssignment, p: P) -> J:
        """
        Handle chained assignment e.g. a = b = 1 <-> a=b=1
        """
        a: ChainedAssignment = cast(ChainedAssignment, super().visit_chained_assignment(chained_assignment, p))
        a = a.padding.with_variables(
            list_map(lambda v: v.with_after(update_space(v.after, self._style.around_operators.assignment)), a.padding.variables))

        a = a.padding.with_variables(
            list_map(lambda v, idx: v.with_element(
                space_before(v.element, self._style.around_operators.assignment if idx >= 1 else False)), a.padding.variables))

        return a.with_assignment(space_before(a.assignment, self._style.around_operators.assignment))

    def visit_member_reference(self, member_reference: MemberReference, p: P) -> J:
        m: MemberReference = cast(MemberReference, super().visit_member_reference(member_reference, p))

        if m.padding.type_parameters is not None:
            # TODO: Handle type parameters, relevant for constructors in Python.
            raise NotImplementedError("Type parameters are not supported yet")

        return m

    def _apply_binary_space_around(self, binary, use_space_around: bool):
        operator = binary.padding.operator
        binary = binary.padding.with_operator(
            operator.with_before(update_space(operator.before, use_space_around))
        )
        return binary.with_right(space_before(binary.right, use_space_around))

    def visit_binary(self, binary: j.Binary, p: P) -> J:
        b: j.Binary = cast(j.Binary, super().visit_binary(binary, p))
        op: j.Binary.Type = b.operator

        # Handle space around all binary operators e.g. 1 +   1 <-> 1 + 1 or 1+1 depending on style
        if op in [j.Binary.Type.Addition, j.Binary.Type.Subtraction]:
            b = self._apply_binary_space_around(b, self._style.around_operators.additive)
        elif op in [j.Binary.Type.Multiplication, j.Binary.Type.Division, j.Binary.Type.Modulo]:
            b = self._apply_binary_space_around(b, self._style.around_operators.multiplicative)
        elif op in [j.Binary.Type.Equal, j.Binary.Type.NotEqual]:
            b = self._apply_binary_space_around(b, self._style.around_operators.equality)
        elif op in [j.Binary.Type.LessThan, j.Binary.Type.GreaterThan, j.Binary.Type.LessThanOrEqual,
                    j.Binary.Type.GreaterThanOrEqual]:
            b = self._apply_binary_space_around(b, self._style.around_operators.relational)
        elif op in [j.Binary.Type.BitAnd, j.Binary.Type.BitOr, j.Binary.Type.BitXor]:
            b = self._apply_binary_space_around(b, self._style.around_operators.bitwise)
        elif op in [j.Binary.Type.LeftShift, j.Binary.Type.RightShift, j.Binary.Type.UnsignedRightShift]:
            b = self._apply_binary_space_around(b, self._style.around_operators.shift)
        elif op in [j.Binary.Type.Or, j.Binary.Type.And]:
            b = self._apply_binary_space_around(b, True)
        else:
            raise NotImplementedError(f"Operation {op} is not supported yet")
        return b

    def visit_python_binary(self, binary: Binary, p: P) -> J:
        b: Binary = cast(Binary, super().visit_python_binary(binary, p))
        op: Binary.Type = b.operator

        if op == Binary.Type.In or op == Binary.Type.Is or op == Binary.Type.IsNot or op == Binary.Type.NotIn:
            # TODO: Not sure what style options to use for these operators
            b = self._apply_binary_space_around(b, True)
        elif op in [Binary.Type.FloorDivision, Binary.Type.MatrixMultiplication]:
            b = self._apply_binary_space_around(b, self._style.around_operators.multiplicative)
        elif op == Binary.Type.StringConcatenation:
            b = self._apply_binary_space_around(b, self._style.around_operators.additive)
        elif op == Binary.Type.Power:
            b = self._apply_binary_space_around(b, self._style.around_operators.power)

        return b

    def visit_if(self, if_stm: If, p: P) -> J:
        if_: j.If = cast(If, super().visit_if(if_stm, p))

        # Handle space before if colon e.g. if True:    pass <-> if True: pass
        if_ = if_.with_if_condition(
            if_.if_condition.padding.with_tree(
                space_after(if_.if_condition.padding.tree, self._style.other.before_colon))
        )
        return if_

    def visit_else(self, else_: If.Else, p: P) -> J:
        e: j.If.Else = cast(j.If.Else, super().visit_else(else_, p))
        e = e.padding.with_body(space_before_right_padded_element(e.padding.body, self._style.other.before_colon))

        return e

    def visit_slice(self, slice: Slice, p: P) -> J:
        s: Slice = cast(Slice, super().visit_slice(slice, p))
        use_space = self._style.within.brackets

        if s.padding.start is not None:
            s = s.padding.with_start(space_before_right_padded_element(s.padding.start, use_space))
            s = s.padding.with_start(space_after_right_padded(s.padding.start, use_space))
        if s.padding.stop is not None:
            # TODO: Check if "before" is correct, in PyCharm it is not supported but we could still support it.
            s = s.padding.with_stop(space_before_right_padded_element(s.padding.stop, use_space))
            s = s.padding.with_stop(space_after_right_padded(s.padding.stop, use_space))

            if s.padding.step is not None:
                # TODO: Check if "before" is correct, in PyCharm it is not supported but we could still support it.
                s = s.padding.with_step(space_before_right_padded_element(s.padding.step, use_space))
                s = s.padding.with_step(space_after_right_padded(s.padding.step, use_space))
        return s

    def visit_python_for_loop(self, for_loop: ForLoop, p: P) -> J:
        fl = cast(ForLoop, super().visit_python_for_loop(for_loop, p))

        # Set single space before loop target e.g. for    i in...: <-> for i in ...:
        fl = fl.with_target(space_before(fl.target, True))

        # Set single space before loop iterable, and in keyword e.g. for i in    []: <-> for i in []:
        fl = fl.padding.with_iterable(space_before_left_padded(fl.padding.iterable, True))
        fl = fl.padding.with_iterable(space_before_right_padded_element(fl.padding.iterable, True))
        return fl

    def visit_parameterized_type(self, parameterized_type: ParameterizedType, p: P) -> J:
        pt = cast(ParameterizedType, super().visit_parameterized_type(parameterized_type, p))

        def _process_element(index, arg, last, use_space):
            if index == 0:
                arg = arg.with_element(space_before(arg.element, use_space))
            else:
                arg = arg.with_element(space_before(arg.element, self._style.other.after_comma))

            if last:
                arg = space_after(arg, use_space and arg.markers.find_first(j.TrailingComma) is None)
                arg = arg.with_markers(arg.markers.compute_by_type(j.TrailingComma, self._remap_trailing_comma_space))
            else:
                arg = space_after(arg, self._style.other.before_comma)

            return arg

        pt = pt.padding.with_type_parameters(
            pt.padding.type_parameters.padding.with_elements(
                list_map(
                    lambda arg, idx: _process_element(idx, arg,
                                                      last=idx == len(pt.padding.type_parameters.padding.elements) - 1,
                                                      use_space=self._style.within.brackets),
                    pt.padding.type_parameters.padding.elements
                )
            )
        )

        return pt

    def visit_parentheses(self, parentheses: Parentheses, p: P) -> J:
        p2 = cast(Parentheses, super().visit_parentheses(parentheses, p))
        p2 = p2.with_prefix(update_space(p2.prefix, False))
        p2 = p2.padding.with_tree(p2.padding.tree.with_after(update_space(p2.padding.tree.after, False)))
        return p2

    def visit_collection_literal(self, collection_literal: CollectionLiteral, p: P) -> J:
        cl = cast(CollectionLiteral, super().visit_collection_literal(collection_literal, p))

        # TODO: Refactor to remove duplicate code.
        def _process_element(index, arg, args_size, use_space):
            if index == 0:
                arg = arg.with_element(space_before(arg.element, use_space))
            else:
                arg = arg.with_element(space_before(arg.element, self._style.other.after_comma))

            if index == args_size - 1:
                arg = space_after(arg, use_space and arg.markers.find_first(j.TrailingComma) is None)
                arg = arg.with_markers(arg.markers.compute_by_type(j.TrailingComma, self._remap_trailing_comma_space))
            else:
                arg = space_after(arg, self._style.other.before_comma)

            return arg

        if cl.kind == CollectionLiteral.Kind.SET:
            _space_style = self._style.within.braces
        elif cl.kind == CollectionLiteral.Kind.LIST:
            _space_style = self._style.within.brackets
        elif cl.kind == CollectionLiteral.Kind.TUPLE:
            _space_style = self._style.within.brackets if self.cursor.first_enclosing(ExpressionTypeTree) else False

        cl = cl.padding.with_elements(
            cl.padding.elements.padding.with_elements(
                list_map(
                    lambda arg, idx: _process_element(idx, arg,
                                                      args_size=len(cl.padding.elements.padding.elements),
                                                      use_space=_space_style),
                    cl.padding.elements.padding.elements)))

        return cl

    def visit_dict_literal(self, dict_literal: DictLiteral, p: P) -> J:
        dl = cast(DictLiteral, super().visit_dict_literal(dict_literal, p))

        def _process_kv_pair(c: JRightPadded[KeyValue], idx: int, arg_size) -> JRightPadded[KeyValue]:
            # Handle both space before and after each comma in key-value pair and space before/after braces in
            # first argument e.g. {1: 2  ,   3: 4} <-> {1: 2, 3: 4}
            if idx == 0 or idx == arg_size - 1:
                if idx == 0:
                    before_comma_style = self._style.other.before_comma
                else:
                    before_comma_style = self._style.within.braces and c.markers.find_first(j.TrailingComma) is None
                after_comma_style = self._style.within.braces if idx == 0 else self._style.other.after_comma
                c = space_after_right_padded(c, before_comma_style)
                c = space_before_right_padded_element(c, after_comma_style)
            else:
                c = space_after_right_padded(c, self._style.other.before_comma)
                c = space_before_right_padded_element(c, self._style.other.after_comma)

            # Handle trailing comma space for last argument e.g. {1: 2, 3: 4, } <-> {1: 2, 3: 4,}
            if idx == arg_size - 1:
                c = c.with_markers(c.markers.compute_by_type(j.TrailingComma, self._remap_trailing_comma_space))

            return c

        dl = dl.padding.with_elements(
            dl.padding.elements.padding.with_elements(
                list_map(lambda x, idx: _process_kv_pair(x, idx, len(dl.padding.elements.padding.elements)),
                         dl.padding.elements.padding.elements)
            )
        )
        return dl

    def visit_multi_import(self, multi_import: MultiImport, p: P) -> J:
        mi: MultiImport = cast(MultiImport, super().visit_multi_import(multi_import, p))

        # Space after from
        if mi.padding.from_ is not None:
            mi = mi.padding.with_from(
                space_before_right_padded_element(mi.padding.from_, True)
            )

        # Single space after import keyword
        mi = mi.padding.with_names(
            space_before_container(mi.padding.names, True)
        )

        _names = mi.padding.names

        # Will handle space after import name before comma e.g. import foo   , bar <-> import foo, bar
        _names = _names.padding.with_elements(
            list_map(lambda x, idx: space_after_right_padded(x, self._style.other.before_comma),
                     _names.padding.elements)
        )
        _names = _names.padding.with_elements(
            list_map(lambda x, idx: space_before_right_padded_element(x, self._style.other.after_comma and idx != 0),
                     _names.padding.elements)
        )

        return mi.padding.with_names(_names)

    def visit_import(self, import_: Import, p: P) -> J:
        imp: Import = cast(Import, super().visit_import(import_, p))
        # Always use single space before and after alias 'as' keyword e.g. import foo  as  bar <-> import foo as bar
        if imp.padding.alias:
            imp = imp.with_alias(space_before(imp.alias, True))
            imp = imp.padding.with_alias(space_before_left_padded(imp.padding.alias, True))
        return imp

    def visit_type_hint(self, type_hint: TypeHint, p: P) -> J:
        th: TypeHint = cast(TypeHint, super().visit_type_hint(type_hint, p))
        th = space_before(th, self._style.other.before_colon)
        th = th.with_type_tree(space_before(th.type_tree, self._style.other.after_colon))
        return th

    def visit_expression_type_tree(self, expression_type_tree: ExpressionTypeTree, p: P) -> J:
        ett = cast(ExpressionTypeTree, super().visit_expression_type_tree(expression_type_tree, p))
        # Can always be set to false as the collection literal will handle the space before the brackets
        ett = space_before(ett, False)
        return ett

    def visit_comprehension_expression(self, comprehension_expression: ComprehensionExpression, p: P) -> J:
        ce = cast(ComprehensionExpression, super().visit_comprehension_expression(comprehension_expression, p))

        # Handle space before result this will depend on the style setting for the comprehension type.
        if ce.kind == ComprehensionExpression.Kind.LIST:
            ce = ce.with_result(space_before(ce.result, self._style.within.brackets))
            ce = ce.with_suffix(update_space(ce.suffix, self._style.within.brackets))
        elif ce.kind == ComprehensionExpression.Kind.GENERATOR:
            ce = ce.with_result(space_before(ce.result, False))
            ce = ce.with_suffix(update_space(ce.suffix, False))
        elif ce.kind in (ComprehensionExpression.Kind.SET, ComprehensionExpression.Kind.DICT):
            ce = ce.with_result(space_before(ce.result, self._style.within.braces))
            ce = ce.with_suffix(update_space(ce.suffix, self._style.within.braces))

        return ce

    def visit_key_value(self, key_value: KeyValue, p: P) -> J:
        # Handle space before and after colon in key-value pair e.g. {1 :   2} <-> {1: 2}
        kv = cast(KeyValue, super().visit_key_value(key_value, p))
        kv = kv.with_value(space_before(kv.value, self._style.other.after_colon))
        return kv.padding.with_key(space_after_right_padded(kv.padding.key, self._style.other.before_colon))

    def visit_comprehension_condition(self, condition: ComprehensionExpression.Condition, p: P) -> J:
        cond = cast(ComprehensionExpression.Condition, super().visit_comprehension_condition(condition, p))
        # Set single space before and after comprehension 'if' keyword.
        cond = space_before(cond, True)
        cond = cond.with_expression(space_before(cond.expression, True))
        return cond

    def visit_comprehension_clause(self, clause: ComprehensionExpression.Clause, p: P) -> J:
        cc = cast(ComprehensionExpression.Clause, super().visit_comprehension_clause(clause, p))

        # Ensure single space before 'for' keyword
        cc = space_before(cc, True)

        # Single before 'in' keyword e.g. ..i   in...  <-> ...i in...
        cc = cc.padding.with_iterated_list(space_before_left_padded(cc.padding.iterated_list, True))
        # Single space before 'iterator' variable (or after for keyword) e.g. ...for   i <-> ...for i
        cc = cc.with_iterator_variable(space_before(cc.iterator_variable, True))
        # Ensure single space after 'in' keyword e.g. ...in   range(10) <-> ...in range(10)
        cc = cc.padding.with_iterated_list(space_before_left_padded_element(cc.padding.iterated_list, True))
        return cc

    def _remap_trailing_comma_space(self, tc: j.TrailingComma) -> j.TrailingComma:
        return tc.with_suffix(update_space(tc.suffix, self._style.other.after_comma))


J2 = TypeVar('J2', bound=j.J)


def space_before(j: J2, add_space: bool) -> J2:
    prefix: Space = cast(Space, j.prefix)

    if prefix.comments or ('\n' in prefix.whitespace):
        return j

    if add_space and not_single_space(prefix.whitespace):
        return j.with_prefix(prefix.with_whitespace(" "))  # pyright: ignore [reportReturnType]
    elif not add_space and only_spaces_and_not_empty(prefix.whitespace):
        return j.with_prefix(prefix.with_whitespace(""))  # pyright: ignore [reportReturnType]

    return j


def space_before_container(container: j.JContainer, add_space: bool) -> j.JContainer:
    if container.before.comments:
        # Perform the space rule for the suffix of the last comment only. Same as IntelliJ.
        comments: List[j.Comment] = space_last_comment_suffix(container.before.comments, add_space)
        return container.with_before(container.before.with_comments(comments))

    if add_space and not_single_space(container.before.whitespace):
        return container.with_before(container.before.with_whitespace(" "))
    elif not add_space and only_spaces_and_not_empty(container.before.whitespace):
        return container.with_before(container.before.with_whitespace(""))
    else:
        return container


def space_before_left_padded_element(left: j.JLeftPadded, add_space: bool) -> j.JLeftPadded:
    return left.with_element(space_before(left.element, add_space))

def space_before_right_padded_element(right: j.JRightPadded, add_space: bool) -> j.JRightPadded:
    return right.with_element(space_before(right.element, add_space))


def space_last_comment_suffix(comments: List[j.Comment], add_space: bool) -> List[j.Comment]:
    return list_map(lambda c, i: space_suffix(c, add_space) if i == len(comments) - 1 else c, comments)


def space_suffix(comment: j.Comment, add_space: bool) -> j.Comment:
    if add_space and not_single_space(comment.suffix):
        return comment.with_suffix(" ")
    elif not add_space and only_spaces_and_not_empty(comment.suffix):
        return comment.with_suffix("")
    else:
        return comment


def space_before_left_padded(j: JLeftPadded[J2], add_space) -> JLeftPadded[J2]:
    space: Space = cast(Space, j.before)
    if space.comments or '\\' in space.whitespace:
        # don't touch whitespaces with comments or continuation characters
        return j

    if add_space and not_single_space(space.whitespace):
        return j.with_before(space.with_whitespace(" "))
    elif not add_space and only_spaces_and_not_empty(space.whitespace):
        return j.with_before(space.with_whitespace(""))
    return j


def space_after(j: J2, add_space: bool) -> J2:
    space: Space = cast(Space, j.after)
    if space.comments or '\\' in space.whitespace:
        # don't touch whitespaces with comments or continuation characters
        return j

    if add_space and not_single_space(space.whitespace):
        return j.with_after(space.with_whitespace(" "))
    elif not add_space and only_spaces_and_not_empty(space.whitespace):
        return j.with_after(space.with_whitespace(""))
    return j


def space_after_right_padded(right: JRightPadded[J2], add_space: bool) -> JRightPadded[J2]:
    space: Space = right.after
    if space.comments or '\\' in space.whitespace:
        # don't touch whitespaces with comments or continuation characters
        return right

    if add_space and not_single_space(space.whitespace):
        return right.with_after(space.with_whitespace(" "))
    elif not add_space and only_spaces_and_not_empty(space.whitespace):
        return right.with_after(space.with_whitespace(""))
    return right


def update_space(s: Space, have_space: bool) -> Space:
    if s.comments:
        return s

    if have_space and not_single_space(s.whitespace):
        return s.with_whitespace(" ")
    elif not have_space and only_spaces_and_not_empty(s.whitespace):
        return s.with_whitespace("")
    else:
        return s


def only_spaces(s: Optional[str]) -> bool:
    return s is not None and all(c in {' ', '\t'} for c in s)


def only_spaces_and_not_empty(s: Optional[str]) -> bool:
    return bool(s) and only_spaces(s)


def not_single_space(s: Optional[str]) -> bool:
    return s is not None and only_spaces(s) and s != " "



from __future__ import annotations

import re
from dataclasses import dataclass, field
from typing import Any, Callable, Optional, cast, List, Type

from . import CompilationUnit, ExpressionStatement, PythonVisitor
from .parser import PythonParserBuilder
from .printer import PythonPrinter
from .. import PrintOutputCapture, Tree
from ..java import J, JavaCoordinates, Space, Identifier, P, Literal, Expression, Statement, Block
from ..parser import ParserBuilder
from ..utils import list_flat_map
from ..visitor import Cursor


@dataclass
class PythonTemplate:
    code: str
    on_after_variable_substitution: Optional[Callable[[str], None]] = None
    _template_parser: PythonTemplateParser = field(init=False, repr=False)

    def __init__(self, code: str, parser_builder: PythonParserBuilder = PythonParserBuilder(),
                 on_after_variable_substitution: Optional[Callable[[str], None]] = None):
        self.code = code
        self.on_after_variable_substitution = on_after_variable_substitution
        self._template_parser = PythonTemplateParser(False, parser_builder, self.on_after_variable_substitution)

    def apply(self, scope: Cursor, coordinates: JavaCoordinates, *parameters) -> J:
        substitutions = self.substitutions(*parameters)
        substituted = substitutions.substitute()
        if self.on_after_variable_substitution:
            self.on_after_variable_substitution(substituted)

        return PythonTemplatePythonExtension(self._template_parser, substitutions, substituted, coordinates)\
            .visit(cast(Tree, scope.value), 0, scope.parent_or_throw)

    def substitutions(self, parameters: List[Any]) -> Substitutions:
        return Substitutions(self.code, parameters)


@dataclass
class PythonTemplateParser:
    context_sensitive: bool
    parser_builder: ParserBuilder
    on_after_variable_substitution: Optional[Callable[[str], None]] = None

    def parse_expression(self, scope: Cursor, template: str, loc: Space.Location) -> J:
        cu: CompilationUnit = next(iter(self.parser_builder.build().parse_strings(template)))
        j = cast(ExpressionStatement, cu.statements[0]).expression if isinstance(cu.statements[0],
                                                                                 ExpressionStatement) else \
            cu.statements[0]
        return j.with_prefix(cast(J, scope.value).prefix)

    def parse_block_statements(self, cursor: Cursor, expected: Type, template: str, loc: Space.Location, mode: JavaCoordinates.Mode) -> List[J]:
        cu: CompilationUnit = next(iter(self.parser_builder.build().parse_strings(template)))
        return cu.statements


@dataclass
class Substitutions:
    code: str
    parameters: List[Any]

    def substitute(self) -> str:
        result = ""
        pos = 0
        param_count = 0
        named_index = {}

        while True:
            # Find next #{
            start = self.code.find('#{', pos)
            if start == -1:
                # No more parameters, add remaining code
                result += self.code[pos:]
                break

            # Add code before the #{
            result += self.code[pos:start]

            # Find matching }
            brace_count = 1
            i = start + 2  # Skip past #{
            while i < len(self.code) and brace_count > 0:
                if self.code[i] == '{':
                    brace_count += 1
                elif self.code[i] == '}':
                    brace_count -= 1
                i += 1

            if brace_count > 0:
                raise ValueError("Unmatched { in template")
            pattern = self.code[start + 2:i - 1].strip()
            param_index = param_count if not pattern or pattern.find('any()') >= 0 else named_index[pattern]
            if (sep := pattern.find(':')) >= 0:
                named_index[pattern[0:sep]] = param_index

            # Replace the #{...} with parameter placeholder
            result += f'__p{param_index}__' if pattern else self.substitute_untyped(param_index)
            if not pattern or pattern.find('any()') >= 0:
                param_count += 1

            pos = i

        return result

    def substitute_untyped(self, index: int) -> str:
        param = self.parameters[index]
        if isinstance(param, Literal):
            capture = PrintOutputCapture(0)
            PythonPrinter().visit(param, capture)
            return capture.get_out()
        return str(param)

    def unsubstitute(self, parsed: J) -> J:
        return cast(J, UnsubstitutionVisitor(self.parameters).visit(parsed, 0))

    def unsubstitute_all(self, parsed: List[J]) -> List[J]:
        return [self.unsubstitute(j) for j in parsed]


@dataclass
class UnsubstitutionVisitor(PythonVisitor[int]):
    parameters: List[Any]
    _param_pattern = re.compile('__p(\d+)__')

    def visit_identifier(self, identifier: Identifier, p: P) -> J:
        if match := self._param_pattern.fullmatch(identifier.simple_name):
            return cast(J, self.parameters[int(match.group(1))]).with_prefix(identifier.prefix)
        return identifier

@dataclass
class PythonTemplatePythonExtension(PythonVisitor[int]):
    template_parser: PythonTemplateParser
    substitutions: Substitutions
    substituted_template: str
    coordinates: JavaCoordinates

    def __post_init__(self):
        self.insertion_point = self.coordinates.tree
        self.loc = self.coordinates.loc
        self.mode = self.coordinates.mode

    def visit_expression(self, expression: Expression, p: int) -> J:
        if (self.loc == Space.Location.EXPRESSION_PREFIX or self.loc == Space.Location.STATEMENT_PREFIX and
            isinstance(expression, Statement)) and expression.is_scope(self.insertion_point):
            parsed = self.template_parser.parse_expression(self.cursor, self.substituted_template, self.loc)
            return self.auto_format(self.substitutions.unsubstitute(parsed).with_prefix(expression.prefix), p)
        return expression

    def visit_block(self, block: Block, p: P) -> J:
        if self.loc == Space.Location.BLOCK_END and block.is_scope(self.insertion_point):
            parsed = self.template_parser.parse_block_statements(Cursor(self.cursor, self.insertion_point), Statement,
                                                                 self.substituted_template, self.loc, self.mode)
            gen: List[Statement] = self.substitutions.unsubstitute_all(parsed)
            return self.auto_format(block.with_statements(block.statements + gen), p, self.cursor.parent) if gen else block
        if self.loc == Space.Location.STATEMENT_PREFIX:
            return self.auto_format(block.with_statements(list_flat_map(lambda s: self.get_replacements(s) if s.is_scope(self.insertion_point) else s, block.statements)), p, self.cursor.parent)
        return super().visit_block(block, p)

    def visit_statement(self, statement: Statement, p: P) -> J:
        return statement
        # if (self.loc == Space.Location.STATEMENT_PREFIX and statement.is_scope(self.insertion_point):
        #     parsed = self.template_parser.parse_expression(self.cursor, self.substituted_template, self.loc)
        #     return self.auto_format(self.substitutions.unsubstitute(parsed).with_prefix(expression.prefix), p)
        # return expression

    def get_replacements(self, statement: Statement) -> List[J]:
        parsed = self.template_parser.parse_block_statements(Cursor(self.cursor, self.insertion_point), Statement, self.substituted_template, self.loc, self.mode)
        gen = self.substitutions.unsubstitute_all(parsed)
        formatted = [s.with_prefix(statement.prefix.with_comments([])) for s in gen]

        if self.mode == JavaCoordinates.Mode.REPLACE:
            return formatted
        elif self.mode == JavaCoordinates.Mode.BEFORE:
            return formatted + [statement]
        else:
            return [statement] + formatted



from typing import Optional, List, cast
from typing import Union, Callable

from rewrite.java import JavaVisitor, Space, JContainer, J, MethodDeclaration, \
    OmitParentheses, Try, JRightPadded, JLeftPadded, VariableDeclarations, Unary, Throw, Ternary, Switch, \
    ControlParentheses, ParameterizedType, ArrayDimension, NewArray, Modifier, MethodInvocation, Literal, \
    Lambda, Import, If, Identifier, ForEachLoop, \
    ClassDeclaration, Case, Block, \
    AssignmentOperation, Assignment, Assert, Annotation, Statement, Loop, Empty, Semicolon, TrailingComma, Yield, \
    WhileLoop, Return, Parentheses, FieldAccess, Continue, \
    Break, Comment, TextComment
from rewrite.java import tree as j
from ..visitor import T
from . import *
from .support_types import P, Py, PyRightPadded, PySpace, PyLeftPadded, PyContainer, J2
from .. import PrintOutputCapture, Cursor, Markers, Tree, Marker
from .visitor import PythonVisitor

JAVA_MARKER_WRAPPER: Callable[[str], str] = lambda out: f"/*~~{out}{'~~' if out else ''}>*/"""


class PythonPrinter(PythonVisitor[PrintOutputCapture[P]]):
    def __init__(self):
        super().__init__()
        self.delegate = PythonJavaPrinter(self)

    def visit(self, tree: Optional[T], p: PrintOutputCapture[P], parent: Optional[Cursor] = None) -> Optional[J]:
        if tree is None:
            return cast(Optional[J], self.default_value(None, p))
        if not isinstance(tree, Py):
            return self.delegate.visit(tree, p, parent)
        else:
            return super().visit(tree, p, parent)

    @PythonVisitor.cursor.setter
    def cursor(self, cursor: Cursor) -> None:
        self.delegate._cursor = cursor
        self._cursor = cursor

    def visit_async(self, async_: Async, p: PrintOutputCapture[P]) -> J:
        self.before_syntax(async_, PySpace.Location.ASYNC_PREFIX, p)
        p.append("async")
        self.visit(async_.statement, p)
        return async_

    def visit_await(self, await_: Await, p: PrintOutputCapture[P]) -> J:
        self.before_syntax(await_, PySpace.Location.AWAIT_PREFIX, p)
        p.append("await")
        self.visit(await_.expression, p)
        return await_

    def visit_python_binary(self, binary: Binary, p: PrintOutputCapture[P]) -> J:
        self.before_syntax(binary, PySpace.Location.BINARY_PREFIX, p)
        self.visit(binary.left, p)
        self.visit_space(binary.padding.operator.before, PySpace.Location.BINARY_OPERATOR, p)

        if binary.operator == Binary.Type.NotIn:
            p.append("not")
            if binary.negation is not None:
                self.visit_space(binary.negation, PySpace.Location.BINARY_NEGATION, p)
            else:
                p.append(' ')
            p.append("in")
        elif binary.operator == Binary.Type.In:
            p.append("in")
        elif binary.operator == Binary.Type.Is:
            p.append("is")
        elif binary.operator == Binary.Type.IsNot:
            p.append("is")
            if binary.negation is not None:
                self.visit_space(binary.negation, PySpace.Location.BINARY_NEGATION, p)
            else:
                p.append(' ')
            p.append("not")
        elif binary.operator == Binary.Type.FloorDivision:
            p.append("//")
        elif binary.operator == Binary.Type.MatrixMultiplication:
            p.append("@")
        elif binary.operator == Binary.Type.Power:
            p.append("**")
        elif binary.operator == Binary.Type.StringConcatenation:
            pass  # empty
        else:
            raise ValueError(f"Unexpected binary operator: {binary.operator}")

        self.visit(binary.right, p)
        self.after_syntax(binary, p)
        return binary

    def visit_chained_assignment(self, chained_assignment: ChainedAssignment, p: PrintOutputCapture[P]) -> J:
        self.before_syntax(chained_assignment, PySpace.Location.CHAINED_ASSIGNMENT_PREFIX, p)
        # TODO: Check name inconsistency between java and python CHAINED_ASSIGNMENT_VARIABLES vs CHAINED_ASSIGNMENT_VARIABLE
        self._print_right_padded(chained_assignment.padding.variables,
                                 PyRightPadded.Location.CHAINED_ASSIGNMENT_VARIABLES, "=", p)
        p.append('=')
        self.visit(chained_assignment.assignment, p)
        self.after_syntax(chained_assignment, p)
        return chained_assignment

    def visit_exception_type(self, exception_type: ExceptionType, p: PrintOutputCapture[P]) -> J:
        self.before_syntax(exception_type, PySpace.Location.EXCEPTION_TYPE_PREFIX, p)
        if exception_type.exception_group:
            p.append("*")
        self.visit(exception_type.expression, p)
        return exception_type

    def visit_python_for_loop(self, for_loop: ForLoop, p: PrintOutputCapture[P]) -> J:
        self.before_syntax(for_loop, PySpace.Location.FOR_LOOP_PREFIX, p)
        p.append("for")
        self.visit(for_loop.target, p)
        self._print_visit_left_padded("in", for_loop.padding.iterable, PyLeftPadded.Location.FOR_LOOP_ITERABLE, p)
        self.visit_right_padded(for_loop.padding.body, PyRightPadded.Location.FOR_LOOP_BODY, p)
        return for_loop

    def visit_literal_type(self, literal_type: LiteralType, p: PrintOutputCapture[P]) -> J:
        self.before_syntax(literal_type, PySpace.Location.LITERAL_TYPE_PREFIX, p)
        self.visit(literal_type.literal, p)
        self.after_syntax(literal_type, p)
        return literal_type

    def visit_marker(self, marker: Marker, p: PrintOutputCapture[P]) -> Marker:
        if isinstance(marker, Semicolon):
            p.append(';')
        elif isinstance(marker, TrailingComma):
            p.append(',')
            self.visit_space(marker.suffix, Space.Location.ANY, p)
        return marker

    def visit_type_hint(self, type_hint: TypeHint, p: PrintOutputCapture[P]) -> J:
        self.before_syntax(type_hint, PySpace.Location.TYPE_HINT_PREFIX, p)
        parent = self.cursor.parent_tree_cursor().value
        p.append("->" if isinstance(parent, MethodDeclaration) else ":")
        self.visit(type_hint.type_tree, p)
        self.after_syntax(type_hint, p)
        return type_hint

    def visit_compilation_unit(self, cu: CompilationUnit, p: PrintOutputCapture[P]) -> J:
        self.before_syntax(cu, Space.Location.COMPILATION_UNIT_PREFIX, p)
        for import_ in cu.padding.imports:
            self.visit_right_padded(import_, PyRightPadded.Location.TOP_LEVEL_STATEMENT_SUFFIX, p)
        for statement in cu.padding.statements:
            self.visit_right_padded(statement, PyRightPadded.Location.TOP_LEVEL_STATEMENT_SUFFIX, p)

        self.visit_space(cu.eof, Space.Location.COMPILATION_UNIT_EOF, p)
        # TODO: The SuppressNewline marker used in the PythonPrinter does not exist?
        # if any(isinstance(marker, SuppressNewline) for marker in cu.markers.markers):
        #     out = p.get_out()
        #     if out and out[-1] == '\n':
        #         p._out[-1] = out[:-1]

        self.after_syntax(cu, p)
        return cu

    def visit_multi_import(self, multi_import: MultiImport, p: PrintOutputCapture[P]) -> J:
        self.before_syntax(multi_import, PySpace.Location.MULTI_IMPORT_PREFIX, p)
        if multi_import.from_ is not None:
            p.append("from")
            self.visit_right_padded(multi_import.padding.from_, PyRightPadded.Location.MULTI_IMPORT_FROM, p)
        p.append("import")

        if multi_import.parenthesized:
            self._print_container("(", multi_import.padding.names, PyContainer.Location.MULTI_IMPORT_NAMES, ",",
                                          ")", p)
        else:
            self._print_container("", multi_import.padding.names, PyContainer.Location.MULTI_IMPORT_NAMES, ",",
                                          "", p)

        self.after_syntax(multi_import, p)
        return multi_import

    def visit_key_value(self, key_value: KeyValue, p: PrintOutputCapture[P]) -> J:
        self.before_syntax(key_value, PySpace.Location.KEY_VALUE_PREFIX, p)
        self.visit_right_padded(key_value.padding.key, PyRightPadded.Location.KEY_VALUE_KEY_SUFFIX, p)
        p.append(':')
        self.visit(key_value.value, p)
        self.after_syntax(key_value, p)
        return key_value

    def visit_dict_literal(self, dict_literal: DictLiteral, p: PrintOutputCapture[P]) -> J:
        self.before_syntax(dict_literal, PySpace.Location.DICT_LITERAL_PREFIX, p)
        self._print_container("{", dict_literal.padding.elements, PyContainer.Location.DICT_LITERAL_ELEMENTS,
                                      ",", "}",
                              p)
        self.after_syntax(dict_literal, p)
        return dict_literal

    def visit_collection_literal(self, collection_literal: CollectionLiteral, p: PrintOutputCapture[P]) -> J:
        self.before_syntax(collection_literal, PySpace.Location.COLLECTION_LITERAL_PREFIX, p)
        elements = collection_literal.padding.elements

        if collection_literal.kind == CollectionLiteral.Kind.LIST:
            self._print_container("[", elements, PyContainer.Location.COLLECTION_LITERAL_ELEMENTS, ",", "]", p)
        elif collection_literal.kind == CollectionLiteral.Kind.SET:
            self._print_container("{", elements, PyContainer.Location.COLLECTION_LITERAL_ELEMENTS, ",", "}", p)
        elif collection_literal.kind == CollectionLiteral.Kind.TUPLE:
            if elements.markers.find_first(OmitParentheses) is not None:
                self._print_container("", elements, PyContainer.Location.COLLECTION_LITERAL_ELEMENTS, ",", "",
                                      p)
            else:
                self._print_container("(", elements, PyContainer.Location.COLLECTION_LITERAL_ELEMENTS, ",", ")",
                                      p)

        self.after_syntax(collection_literal, p)
        return collection_literal

    def visit_formatted_string(self, formatted_string: FormattedString, p: PrintOutputCapture[P]) -> J:
        self.before_syntax(formatted_string, PySpace.Location.FORMATTED_STRING_PREFIX, p)
        p.append(formatted_string.delimiter)
        for part in formatted_string.parts:
            self.visit(part, p)
        if formatted_string.delimiter:
            idx = max(formatted_string.delimiter.find("'"), formatted_string.delimiter.find('"'))
            p.append(formatted_string.delimiter[idx:])
        return formatted_string

    def visit_formatted_string_value(self, value: FormattedString.Value, p: PrintOutputCapture[P]) -> J:
        self.before_syntax(value, PySpace.Location.FORMATTED_STRING_VALUE_PREFIX, p)
        p.append('{')
        self.visit_right_padded(value.padding.expression, PyRightPadded.Location.FORMATTED_STRING_VALUE_EXPRESSION, p)

        if value.padding.debug is not None:
            p.append('=')
            self.visit_space(value.padding.debug.after, PySpace.Location.FORMATTED_STRING_VALUE_DEBUG_SUFFIX, p)

        if value.conversion is not None:
            p.append('!')
            if value.conversion == FormattedString.Value.Conversion.STR:
                p.append('s')
            elif value.conversion == FormattedString.Value.Conversion.REPR:
                p.append('r')
            elif value.conversion == FormattedString.Value.Conversion.ASCII:
                p.append('a')

        if value.format is not None:
            p.append(':')
            self.visit(value.format, p)

        p.append('}')
        return value

    def visit_pass(self, pass_: Pass, p: PrintOutputCapture[P]) -> J:
        self.before_syntax(pass_, PySpace.Location.PASS_PREFIX, p)
        p.append("pass")
        self.after_syntax(pass_, p)
        return pass_

    def visit_trailing_else_wrapper(self, trailing_else_wrapper: TrailingElseWrapper, p: PrintOutputCapture[P]) -> J:
        self.before_syntax(trailing_else_wrapper, PySpace.Location.TRAILING_ELSE_WRAPPER_PREFIX, p)
        self.visit(trailing_else_wrapper.statement, p)
        if not isinstance(trailing_else_wrapper.statement, Try):
            self.visit_space(
                trailing_else_wrapper.padding.else_block.before,
                Space.Location.ELSE_PREFIX,
                p
            )
            p.append("else")
            self.visit(trailing_else_wrapper.else_block, p)
        self.after_syntax(trailing_else_wrapper, p)
        return trailing_else_wrapper

    def visit_comprehension_expression(self, comprehension_expression: ComprehensionExpression,
                                       p: PrintOutputCapture[P]) -> J:
        self.before_syntax(comprehension_expression, PySpace.Location.COMPREHENSION_EXPRESSION_PREFIX, p)

        if comprehension_expression.kind in {ComprehensionExpression.Kind.DICT, ComprehensionExpression.Kind.SET}:
            open_, close = "{", "}"
        elif comprehension_expression.kind == ComprehensionExpression.Kind.LIST:
            open_, close = "[", "]"
        elif comprehension_expression.kind == ComprehensionExpression.Kind.GENERATOR:
            if comprehension_expression.markers.find_first(OmitParentheses) is not None:
                open_, close = "", ""
            else:
                open_, close = "(", ")"
        else:
            raise ValueError(f"Unexpected comprehension kind: {comprehension_expression.kind}")

        p.append(open_)
        self.visit(comprehension_expression.result, p)
        for clause in comprehension_expression.clauses:
            self.visit(clause, p)

        self.visit_space(comprehension_expression.suffix, PySpace.Location.COMPREHENSION_EXPRESSION_PREFIX, p)
        p.append(close)

        self.after_syntax(comprehension_expression, p)
        return comprehension_expression

    def visit_comprehension_condition(self, condition: ComprehensionExpression.Condition,
                                      p: PrintOutputCapture[P]) -> J:
        self.before_syntax(condition, PySpace.Location.COMPREHENSION_EXPRESSION_CONDITION_PREFIX, p)
        p.append("if")
        self.visit(condition.expression, p)
        return condition

    def visit_comprehension_clause(self, clause: ComprehensionExpression.Clause, p: PrintOutputCapture[P]) -> J:
        self.before_syntax(clause, PySpace.Location.COMPREHENSION_EXPRESSION_CLAUSE_PREFIX, p)
        if clause.async_ and clause.padding.async_ is not None:
            p.append("async")
            self.visit_space(clause.padding.async_.after, PySpace.Location.COMPREHENSION_EXPRESSION_CLAUSE_ASYNC_SUFFIX,
                             p)
        p.append("for")
        self.visit(clause.iterator_variable, p)
        self.visit_space(clause.padding.iterated_list.before, PySpace.Location.COMPREHENSION_IN, p)
        p.append("in")
        self.visit(clause.iterated_list, p)
        if clause.conditions is not None:
            for condition in clause.conditions:
                self.visit(condition, p)
        return clause

    def visit_type_alias(self, type_alias: TypeAlias, p: PrintOutputCapture[P]) -> J:
        self.before_syntax(type_alias, PySpace.Location.UNION_TYPE_PREFIX, p)
        p.append("type")
        self.visit(type_alias.name, p)
        self._print_visit_left_padded("=", type_alias.padding.value, PyLeftPadded.Location.TYPE_ALIAS_VALUE, p)
        self.after_syntax(type_alias, p)
        return type_alias

    def visit_yield_from(self, yield_from: YieldFrom, p: PrintOutputCapture[P]) -> J:
        self.before_syntax(yield_from, PySpace.Location.YIELD_FROM_PREFIX, p)
        p.append("from")
        self.visit(yield_from.expression, p)
        return yield_from

    def visit_union_type(self, union_type: UnionType, p: PrintOutputCapture[P]) -> J:
        self.before_syntax(union_type, PySpace.Location.UNION_TYPE_PREFIX, p)
        # TODO: Check name inconsistency between java and python UNION_TYPE_TYPES vs UNION_TYPE_TYPE
        self._print_right_padded(union_type.padding.types, PyRightPadded.Location.UNION_TYPE_TYPES, "|", p)
        self.after_syntax(union_type, p)
        return union_type

    def visit_variable_scope(self, variable_scope: VariableScope, p: PrintOutputCapture[P]) -> J:
        self.before_syntax(variable_scope, PySpace.Location.VARIABLE_SCOPE_PREFIX, p)
        if VariableScope.Kind.GLOBAL == variable_scope.kind:
            p.append("global")
        elif VariableScope.Kind.NONLOCAL == variable_scope.kind:
            p.append("nonlocal")

        # TODO: Check name inconsistency between java and python VARIABLE_SCOPE_ELEMENTS vs VARIABLE_SCOPE_NAME_SUFFIX
        self._print_right_padded(variable_scope.padding.names, PyRightPadded.Location.VARIABLE_SCOPE_NAMES, ",",
                                 p)
        return variable_scope

    def visit_del(self, del_: Del, p: PrintOutputCapture[P]) -> J:
        self.before_syntax(del_, PySpace.Location.DEL_PREFIX, p)
        p.append("del")
        # Note: PythonPrinter.java uses PyRightPadded.Location.DEL_ELEMENTS here but that doesn't exist in Python.
        # Instead we use PyRightPadded.Location.DEL_TARGETS as both actually point to PySpace.Location.DEL_TARGET_SUFFIX.
        self._print_right_padded(del_.padding.targets, PyRightPadded.Location.DEL_TARGETS, ",", p)
        return del_

    def visit_special_parameter(self, special_parameter: SpecialParameter, p: PrintOutputCapture[P]) -> J:
        self.before_syntax(special_parameter, PySpace.Location.SPECIAL_PARAMETER_PREFIX, p)
        if special_parameter.kind == SpecialParameter.Kind.ARGS:
            p.append("*")
        elif special_parameter.kind == SpecialParameter.Kind.KWARGS:
            p.append("**")
        self.after_syntax(special_parameter, p)
        return special_parameter

    def visit_star(self, star: Star, p: PrintOutputCapture[P]) -> J:
        self.before_syntax(star, PySpace.Location.STAR_PREFIX, p)
        if Star.Kind.LIST == star.kind:
            p.append('*')
        elif Star.Kind.DICT == star.kind:
            p.append('**')
        self.visit(star.expression, p)
        self.after_syntax(star, p)
        return star

    def visit_named_argument(self, named_argument: NamedArgument, p: PrintOutputCapture[P]) -> J:
        self.before_syntax(named_argument, PySpace.Location.NAMED_ARGUMENT, p)
        self.visit(named_argument.name, p)
        self._print_visit_left_padded("=", named_argument.padding.value, PyLeftPadded.Location.NAMED_ARGUMENT_VALUE,
                                      p)
        return named_argument

    def visit_type_hinted_expression(self, type_hinted_expression: TypeHintedExpression, p: PrintOutputCapture[P]) -> J:
        self.before_syntax(type_hinted_expression, PySpace.Location.TYPE_HINTED_EXPRESSION_PREFIX, p)
        self.visit(type_hinted_expression.expression, p)
        self.visit(type_hinted_expression.type_hint, p)
        self.after_syntax(type_hinted_expression, p)
        return type_hinted_expression

    def visit_error_from(self, error_from: ErrorFrom, p: PrintOutputCapture[P]) -> J:
        self.before_syntax(error_from, PySpace.Location.ERROR_FROM_PREFIX, p)
        self.visit(error_from.error, p)
        # TODO: Check name inconsistency between java and python ERROR_FROM_SOURCE vs ERROR_FROM_PREFIX?
        self.visit_space(error_from.padding.from_.before, PySpace.Location.ERROR_FROM_PREFIX, p)
        p.append("from")
        self.visit(error_from.from_, p)
        return error_from

    def visit_match_case(self, match_case: MatchCase, p: PrintOutputCapture[P]) -> J:
        self.before_syntax(match_case, PySpace.Location.MATCH_CASE_PREFIX, p)
        self.visit(match_case.pattern, p)
        if match_case.padding.guard is not None:
            self.visit_space(match_case.padding.guard.before, PySpace.Location.MATCH_CASE_GUARD, p)
            p.append("if")
            self.visit(match_case.guard, p)
        return match_case

    def visit_match_case_pattern(self, pattern: MatchCase.Pattern, p: PrintOutputCapture[P]) -> J:
        self.before_syntax(pattern, PySpace.Location.MATCH_PATTERN_PREFIX, p)
        children = pattern.padding.children
        if pattern.kind == MatchCase.Pattern.Kind.AS:
            self._print_container("", children, PyContainer.Location.MATCH_CASE_PATTERN_CHILDREN, "as", "", p)
        elif pattern.kind in {MatchCase.Pattern.Kind.CAPTURE, MatchCase.Pattern.Kind.LITERAL}:
            self.visit_container(children, PyContainer.Location.MATCH_CASE_PATTERN_CHILDREN, p)
        elif pattern.kind == MatchCase.Pattern.Kind.CLASS:
            self.visit_space(children.before, PySpace.Location.MATCH_CASE_PATTERN_PREFIX, p)
            self.visit_right_padded(children.padding.elements[0], PyRightPadded.Location.MATCH_CASE_PATTERN_CHILD, p)
            self._print_container("(", JContainer(Space.EMPTY, children.padding.elements[1:], Markers.EMPTY),
                                  PyContainer.Location.MATCH_CASE_PATTERN_CHILDREN, ",", ")", p)
        elif pattern.kind == MatchCase.Pattern.Kind.DOUBLE_STAR:
            self._print_container("**", children, PyContainer.Location.MATCH_CASE_PATTERN_CHILDREN, "", "", p)
        elif pattern.kind == MatchCase.Pattern.Kind.KEY_VALUE:
            self._print_container("", children, PyContainer.Location.MATCH_CASE_PATTERN_CHILDREN, ":", "", p)
        elif pattern.kind == MatchCase.Pattern.Kind.KEYWORD:
            self._print_container("", children, PyContainer.Location.MATCH_CASE_PATTERN_CHILDREN, "=", "", p)
        elif pattern.kind == MatchCase.Pattern.Kind.MAPPING:
            self._print_container("{", children, PyContainer.Location.MATCH_CASE_PATTERN_CHILDREN, ",", "}", p)
        elif pattern.kind == MatchCase.Pattern.Kind.OR:
            self._print_container("", children, PyContainer.Location.MATCH_CASE_PATTERN_CHILDREN, "|", "", p)
        elif pattern.kind == MatchCase.Pattern.Kind.SEQUENCE:
            self._print_container("", children, PyContainer.Location.MATCH_CASE_PATTERN_CHILDREN, ",", "", p)
        elif pattern.kind == MatchCase.Pattern.Kind.SEQUENCE_LIST:
            self._print_container("[", children, PyContainer.Location.MATCH_CASE_PATTERN_CHILDREN, ",", "]", p)
        elif pattern.kind in {MatchCase.Pattern.Kind.GROUP, MatchCase.Pattern.Kind.SEQUENCE_TUPLE}:
            self._print_container("(", children, PyContainer.Location.MATCH_CASE_PATTERN_CHILDREN, ",", ")", p)
        elif pattern.kind == MatchCase.Pattern.Kind.STAR:
            self._print_container("*", children, PyContainer.Location.MATCH_CASE_PATTERN_CHILDREN, "", "", p)
        elif pattern.kind == MatchCase.Pattern.Kind.VALUE:
            self._print_container("", children, PyContainer.Location.MATCH_CASE_PATTERN_CHILDREN, "", "", p)
        elif pattern.kind == MatchCase.Pattern.Kind.WILDCARD:
            self._print_container("_", children, PyContainer.Location.MATCH_CASE_PATTERN_CHILDREN, "", "", p)

        return pattern

    def visit_slice(self, slice_: Slice, p: PrintOutputCapture[P]) -> J:
        self.before_syntax(slice_, PySpace.Location.SLICE_PREFIX, p)
        self.visit_right_padded(slice_.padding.start, PyRightPadded.Location.SLICE_START, p)
        p.append(':')
        if slice_.padding.stop is not None:
            self.visit_right_padded(slice_.padding.stop, PyRightPadded.Location.SLICE_STOP, p)
            if slice_.padding.step is not None:
                p.append(':')
                self.visit_right_padded(slice_.padding.step, PyRightPadded.Location.SLICE_STEP, p)
        return slice_

    # CUSTOM VISIT METHODS FOR VISIT CONTAINER, VISIT SPACE, VISIT RIGHT PADDED, VISIT LEFT PADDED
    def _print_container(self, before: str, container: Optional[JContainer[J2]],
                         loc: PyContainer.Location, suffix_between: str, after: str,
                         p: PrintOutputCapture[P]) -> None:
        if container is None:
            return None
        self.visit_space(container.before, loc.before_location, p)
        p.append(before)
        self._print_right_padded(container.padding.elements, loc.element_location, suffix_between, p)
        p.append("" if after is None else after)

    def _print_right_padded(self, nodes: List[JRightPadded[J2]], location: PyRightPadded.Location,
                            suffix_between: str, p: PrintOutputCapture[P]) -> None:
        for i, node in enumerate(nodes):
            self.visit(node.element, p)
            self.visit_space(node.after, location.after_location, p)
            self.visit_markers(node.markers, p)
            if i < len(nodes) - 1:
                p.append(suffix_between)

    def _print_visit_left_padded(self, s: str, left: JLeftPadded[J2], _: PyLeftPadded.Location,
                                 p: PrintOutputCapture[P]) -> None:
        self.delegate.visit_space(left.before, Space.Location.LANGUAGE_EXTENSION, p)
        p.append(s)
        self.cursor = Cursor(self.cursor, left)
        t = left.element
        if isinstance(t, J):
            self.visit_and_cast(left.element, J, p)

        self.cursor = self.cursor.parent  # pyright: ignore [reportAttributeAccessIssue]
        self.visit_markers(left.markers, p)

    def visit_container(self, container: Optional[JContainer[J2]],
                        loc: Union[PyContainer.Location, JContainer.Location], p: PrintOutputCapture[P]) -> JContainer[
        # pyright: ignore
        J2]:
        raise NotImplementedError("Should not be triggered")

    def visit_space(self, space: Optional[Space], loc: Optional[Union[PySpace.Location, Space.Location]],
                    p: PrintOutputCapture[P]) -> Space:
        loc_ = Space.Location.LANGUAGE_EXTENSION if isinstance(loc, PySpace.Location) else loc
        return self.delegate.visit_space(space, loc_, p)  # pyright: ignore [reportArgumentType]

    def before_syntax(
            self,
            py: Py,
            loc: Union[Space.Location, PySpace.Location],
            p: PrintOutputCapture[P],
    ) -> None:
        """
        Unified before_syntax method that handles all cases from Java overloads.
        1. before_syntax(py: Py, loc: Location, p: PrintOutputCapture)
        2. before_syntax(py: Py, loc: PySpace.Location, p: PrintOutputCapture)
        """
        return self._before_syntax_internal(py.prefix, py.markers, loc, p)

    def _before_syntax_internal(
            self,
            prefix: Space,
            markers: Markers,
            loc: Optional[Union[Space.Location, PySpace.Location]],
            p: PrintOutputCapture
    ) -> None:
        """
        Implementation of the common before_syntax logic
        """
        for marker in markers.markers:
            p.append(p.marker_printer.before_prefix(
                marker,
                Cursor(self.cursor, marker),
                JAVA_MARKER_WRAPPER
            ))

        if loc is not None:
            self.visit_space(prefix, loc, p)

        self.visit_markers(markers, p)

        for marker in markers.markers:
            p.append(p.marker_printer.before_syntax(
                marker,
                Cursor(self.cursor, marker),
                JAVA_MARKER_WRAPPER
            ))

    def after_syntax(
            self,
            arg: Union[Py, Markers],
            p: PrintOutputCapture
    ) -> None:
        """
        Unified after_syntax method that handles both cases from Java overloads.
        Usage patterns:
        1. after_syntax(py: Py, p: PrintOutputCapture)
        2. after_syntax(markers: Markers, p: PrintOutputCapture)
        """
        markers = arg.markers if isinstance(arg, Py) else arg

        for marker in markers.markers:
            p.append(p.marker_printer.after_syntax(
                marker,
                Cursor(self.cursor, marker),
                JAVA_MARKER_WRAPPER
            ))


class JavaPrinter(JavaVisitor[PrintOutputCapture[P]]):

    def before_syntax(self,
                      j_or_prefix: Union[J, Space],
                      loc: Optional[Space.Location],
                      p: PrintOutputCapture[P],
                      markers: Markers = Markers.EMPTY) -> None:
        # If j_or_prefix is a J type, extract prefix and markers
        if isinstance(j_or_prefix, J):
            self.before_syntax(j_or_prefix.prefix, loc, p, j_or_prefix.markers)
        else:
            markers_list = markers.markers

            # Process markers before prefix
            for marker in markers_list:
                p.append(p.marker_printer.before_prefix(
                    marker,
                    Cursor(self.cursor, marker),
                    JAVA_MARKER_WRAPPER
                ))

            if loc is not None:
                self.visit_space(j_or_prefix, loc, p)
            self.visit_markers(markers, p)
            for marker in markers_list:
                p.append(p.marker_printer.before_syntax(
                    marker,
                    Cursor(self.cursor, marker),
                    JAVA_MARKER_WRAPPER
                ))

    def after_syntax(self, j_or_markers: Union[J, Markers], p: PrintOutputCapture[P]) -> None:
        # If j_or_markers is a J type, extract markers and call after_syntax
        markers = j_or_markers.markers if isinstance(j_or_markers, J) else j_or_markers

        markers_list = markers.markers
        for marker in markers_list:
            p.append(p.marker_printer.after_syntax(
                marker,
                Cursor(self.cursor, marker),
                JAVA_MARKER_WRAPPER
            ))

    def visit_statements(self, statements: List[JRightPadded[Statement]], location: JRightPadded.Location,
                         p: PrintOutputCapture[P]) -> None:
        for padded_stat in statements:
            self._print_statement(padded_stat, location, p)

    def _print_statement(self, padded_stat: Optional[JRightPadded[Statement]], location: JRightPadded.Location,
                         p: PrintOutputCapture[P]) -> None:
        if padded_stat is None:
            return

        self.visit(padded_stat.element, p)
        self.visit_space(padded_stat.after, location.after_location, p)
        self.visit_markers(padded_stat.markers, p)

    def _print_container(self, before: str, container: Optional[JContainer[J2]], location: JContainer.Location,
                         suffix_between: str, after: Optional[str], p: PrintOutputCapture[P]) -> None:
        if container is None:
            return
        self.before_syntax(container.before, location.before_location, p, markers=container.markers)
        p.append(before)
        self._print_right_padded_list(container.padding.elements, location.element_location, suffix_between, p)
        self.after_syntax(container.markers, p)
        p.append("" if after is None else after)

    def _print_right_padded_list(self, nodes: List[JRightPadded[J2]], location: JRightPadded.Location,
                                 suffix_between: str,
                                 p: PrintOutputCapture[P]) -> None:
        for i, node in enumerate(nodes):
            self.visit(node.element, p)
            self.visit_space(node.after, location.after_location, p)
            self.visit_markers(node.markers, p)
            if i < len(nodes) - 1:
                p.append(suffix_between)

    def _print_right_padded(self, right_padded: Optional[JRightPadded[J2]], location: JRightPadded.Location,
                            suffix: Optional[str], p: PrintOutputCapture[P]) -> None:
        if right_padded is not None:
            self.before_syntax(Space.EMPTY, None, p, right_padded.markers)
            self.visit(right_padded.element, p)
            self.after_syntax(right_padded.markers, p)
            self.visit_space(right_padded.after, location.after_location, p)
            if suffix is not None:
                p.append(suffix)

    def _print_left_padded(self, prefix: Optional[str], left_padded: Optional[JLeftPadded[J2]],
                           location: JLeftPadded.Location, p: PrintOutputCapture[P]) -> None:
        if left_padded is not None:
            self.before_syntax(left_padded.before, location.before_location, p, left_padded.markers)
            if prefix is not None:
                p.append(prefix)
            self.visit(left_padded.element, p)
            self.after_syntax(left_padded.markers, p)

    def visit_space(self, space: Space, _: Optional[Space.Location], p: PrintOutputCapture[P]) -> Space:
        p.append(space.whitespace)
        comments = space.comments
        for comment in comments:
            self.visit_markers(comment.markers, p)
            self.print_comment(comment, p)
            p.append(comment.suffix)
        return space

    @staticmethod
    def print_comment(comment: Comment, p: PrintOutputCapture[P]):
        if isinstance(comment, PyComment):
            p.append("#").append(comment.text)
        elif isinstance(comment, TextComment):
            raise NotImplementedError("TextComment not yet supported")
        else:
            raise ValueError(f"Unexpected comment type: {type(comment)}")

    def visit_field_access(self, field_access: FieldAccess, p: PrintOutputCapture[P]) -> J:
        self.before_syntax(field_access, Space.Location.FIELD_ACCESS_PREFIX, p)
        self.visit(field_access.target, p)
        self._print_left_padded(".", field_access.padding.name, JLeftPadded.Location.FIELD_ACCESS_NAME, p)
        self.after_syntax(field_access, p)
        return field_access

    def visit_parentheses(self, parens: Parentheses[J2], p: PrintOutputCapture[P]) -> J:
        self.before_syntax(parens, Space.Location.PARENTHESES_PREFIX, p)
        p.append('(')
        self._print_right_padded(parens.padding.tree, JRightPadded.Location.PARENTHESES, ")", p)
        self.after_syntax(parens, p)
        return parens

    def visit_while_loop(self, while_loop: WhileLoop, p: PrintOutputCapture[P]) -> J:
        self.before_syntax(while_loop, Space.Location.WHILE_PREFIX, p)
        p.append("while")
        self.visit(while_loop.condition, p)
        self._print_statement(while_loop.padding.body, JRightPadded.Location.WHILE_BODY, p)
        self.after_syntax(while_loop, p)
        return while_loop

    def visit_continue(self, continue_: Continue, p: PrintOutputCapture[P]) -> J:
        self.before_syntax(continue_, Space.Location.CONTINUE_PREFIX, p)
        p.append("continue")
        self.visit(continue_.label, p)
        self.after_syntax(continue_, p)
        return continue_

    def visit_empty(self, empty: Empty, p: PrintOutputCapture[P]) -> J:
        self.before_syntax(empty, Space.Location.EMPTY_PREFIX, p)
        self.after_syntax(empty, p)
        return empty

    def visit_return(self, return_: Return, p: PrintOutputCapture[P]) -> J:
        self.before_syntax(return_, Space.Location.RETURN_PREFIX, p)
        p.append("return")
        self.visit(return_.expression, p)
        self.after_syntax(return_, p)
        return return_

    def visit_yield(self, yield_: Yield, p: PrintOutputCapture[P]) -> Yield:
        self.before_syntax(yield_, Space.Location.YIELD_PREFIX, p)
        if not yield_.implicit:
            p.append("yield")
        self.visit(yield_.value, p)
        self.after_syntax(yield_, p)
        return yield_

    def visit_break(self, break_: Break, p: PrintOutputCapture[P]) -> J:
        self.before_syntax(break_, Space.Location.BREAK_PREFIX, p)
        p.append("break")
        self.visit(break_.label, p)
        self.after_syntax(break_, p)
        return break_

    def visit_left_padded(self, left: Optional[JLeftPadded[T]], loc: JLeftPadded.Location, p: PrintOutputCapture[P]) -> \
            Optional[
        JLeftPadded[T]]:
        return super().visit_left_padded(left, loc, p)

    def visit_right_padded(self, right: Optional[JRightPadded[T]], loc: JRightPadded.Location,
                           p: PrintOutputCapture[P]) -> Optional[
        JRightPadded[T]]:
        return super().visit_right_padded(right, loc, p)


class PythonJavaPrinter(JavaPrinter):
    delegate: PythonPrinter

    def __init__(self, delegate: PythonPrinter):
        self.delegate = delegate

    def visit(self, tree: Union[Optional[Tree], List[J2]], p: PrintOutputCapture[P],
              parent: Optional[Cursor] = None) -> Optional[J]:
        if tree is None:
            return cast(Optional[J], self.default_value(None, p))

        if isinstance(tree, list):
            for t in tree:
                if isinstance(t, Py):
                    self.delegate.visit(t, p)
                else:
                    super().visit(t, p, parent)
        else:
            if isinstance(tree, Py):
                return self.delegate.visit(tree, p)
            else:
                return super().visit(tree, p, parent)

    @property
    def cursor(self) -> Cursor:
        return super().cursor

    @cursor.setter
    def cursor(self, cursor: Cursor) -> None:
        self.delegate._cursor = cursor
        self._cursor = cursor

    def visit_annotation(self, annotation: Annotation, p: PrintOutputCapture[P]) -> J:
        self.before_syntax(annotation, Space.Location.ANNOTATION_PREFIX, p)
        p.append("@")
        self.visit(annotation.annotation_type, p)
        self._print_container("(", annotation.padding.arguments,
                              JContainer.Location.ANNOTATION_ARGUMENTS, ",", ")", p)
        self.after_syntax(annotation, p)
        return annotation

    def visit_array_dimension(self, array_dimension: ArrayDimension, p: PrintOutputCapture[P]) -> J:
        self.before_syntax(array_dimension, Space.Location.DIMENSION_PREFIX, p)
        p.append("[")
        self._print_right_padded(array_dimension.padding.index,
                                 JRightPadded.Location.ARRAY_INDEX, "]", p)
        self.after_syntax(array_dimension, p)
        return array_dimension

    def visit_assert(self, assert_: Assert, p: PrintOutputCapture[P]) -> J:
        self.before_syntax(assert_, Space.Location.ASSERT_PREFIX, p)
        p.append("assert")
        self.visit(assert_.condition, p)
        if assert_.detail is not None:
            self._print_left_padded(",", assert_.detail,
                                    JLeftPadded.Location.ASSERT_DETAIL, p)
        self.after_syntax(assert_, p)
        return assert_

    def visit_assignment(self, assignment: Assignment, p: PrintOutputCapture[P]) -> J:
        parent_tree = self.cursor.parent_tree_cursor().value

        symbol = "=" if (isinstance(parent_tree, (Block, CompilationUnit)) or
                         (isinstance(parent_tree, If) and parent_tree.then_part == assignment) or
                         (isinstance(parent_tree, If.Else) and parent_tree.body == assignment) or
                         (isinstance(parent_tree, Loop) and parent_tree.body == assignment)) else ":="

        self.before_syntax(assignment, Space.Location.ASSIGNMENT_PREFIX, p)
        self.visit(assignment.variable, p)
        self._print_left_padded(symbol, assignment.padding.assignment,
                                JLeftPadded.Location.ASSIGNMENT, p)
        self.after_syntax(assignment, p)
        return assignment

    def visit_assignment_operation(self, assign_op: AssignmentOperation, p: PrintOutputCapture[P]) -> J:
        operator_map = {
            AssignmentOperation.Type.Addition: '+=',
            AssignmentOperation.Type.Subtraction: '-=',
            AssignmentOperation.Type.Multiplication: '*=',
            AssignmentOperation.Type.Division: '/=',
            AssignmentOperation.Type.Modulo: '%=',
            AssignmentOperation.Type.BitAnd: '&=',
            AssignmentOperation.Type.BitOr: '|=',
            AssignmentOperation.Type.BitXor: '^=',
            AssignmentOperation.Type.LeftShift: '<<=',
            AssignmentOperation.Type.RightShift: '>>=',
            AssignmentOperation.Type.UnsignedRightShift: '>>>=',
            AssignmentOperation.Type.Exponentiation: '**=',
            AssignmentOperation.Type.FloorDivision: '//=',
            AssignmentOperation.Type.MatrixMultiplication: '@='
        }

        keyword = operator_map.get(assign_op.operator, None)
        if keyword is None:
            raise ValueError(f"Unknown assignment operator: {assign_op.operator}")

        self.before_syntax(assign_op, Space.Location.ASSIGNMENT_OPERATION_PREFIX, p)
        self.visit(assign_op.variable, p)
        self.visit_space(assign_op.padding.operator.before,
                         Space.Location.ASSIGNMENT_OPERATION_OPERATOR, p)
        p.append(keyword)
        self.visit(assign_op.assignment, p)
        self.after_syntax(assign_op, p)
        return assign_op


    def visit_binary(self, binary: j.Binary, p: PrintOutputCapture[P]) -> J:
        operator_map = {
            j.Binary.Type.Addition: '+',
            j.Binary.Type.Subtraction: '-',
            j.Binary.Type.Multiplication: '*',
            j.Binary.Type.Division: '/',
            j.Binary.Type.Modulo: '%',
            j.Binary.Type.LessThan: '<',
            j.Binary.Type.GreaterThan: '>',
            j.Binary.Type.LessThanOrEqual: '<=',
            j.Binary.Type.GreaterThanOrEqual: '>=',
            j.Binary.Type.Equal: '==',
            j.Binary.Type.NotEqual: '!=',
            j.Binary.Type.BitAnd: '&',
            j.Binary.Type.BitOr: '|',
            j.Binary.Type.BitXor: '^',
            j.Binary.Type.LeftShift: '<<',
            j.Binary.Type.RightShift: '>>',
            j.Binary.Type.UnsignedRightShift: '>>>',
            j.Binary.Type.Or: 'or',
            j.Binary.Type.And: 'and'
        }

        keyword = operator_map.get(binary.operator, None)
        if not keyword:
            raise ValueError(f"Unknown binary operator: {binary.operator}")

        self.before_syntax(binary, Space.Location.BINARY_PREFIX, p)
        self.visit(binary.left, p)
        self.visit_space(binary.padding.operator.before,
                         Space.Location.BINARY_OPERATOR, p)
        p.append(keyword)
        self.visit(binary.right, p)
        self.after_syntax(binary, p)
        return binary

    def visit_block(self, block: Block, p: PrintOutputCapture[P]) -> J:
        self.before_syntax(block, Space.Location.BLOCK_PREFIX, p)
        p.append(':')
        self.visit_statements(block.padding.statements,
                              JRightPadded.Location.BLOCK_STATEMENT, p)
        self.visit_space(block.end, Space.Location.BLOCK_END, p)
        self.after_syntax(block, p)
        return block

    def visit_case(self, ca: Case, p: PrintOutputCapture[P]) -> J:
        self.before_syntax(ca, Space.Location.CASE_PREFIX, p)
        elem = ca.case_labels[0]
        if not (isinstance(elem, Identifier) and elem.simple_name == "default"):
            p.append("case")
        self._print_container("", ca.padding.case_labels,
                              JContainer.Location.CASE_EXPRESSION, ",", "", p)
        self.visit_space(ca.padding.statements.before, Space.Location.CASE, p)
        self.visit_statements(ca.padding.statements.padding.elements,
                              JRightPadded.Location.CASE, p)
        if isinstance(ca.body, Statement):
            self.visit_right_padded(ca.padding.body,
                                    JRightPadded.Location.LANGUAGE_EXTENSION, p)
        else:
            self._print_right_padded(ca.padding.body,
                                     JRightPadded.Location.CASE_BODY, ";", p)
        self.after_syntax(ca, p)
        return ca

    def visit_catch(self, ca: Try.Catch, p: PrintOutputCapture[P]) -> J:
        self.before_syntax(ca, Space.Location.CATCH_PREFIX, p)
        p.append("except")

        multi_variable = ca.parameter.tree
        self.before_syntax(multi_variable, Space.Location.VARIABLE_DECLARATIONS_PREFIX, p)
        self.visit(multi_variable.type_expression, p)

        for padded_variable in multi_variable.padding.variables:
            variable = padded_variable.element
            if variable.name.simple_name == "":
                continue
            self.visit_space(padded_variable.after, Space.Location.LANGUAGE_EXTENSION, p)
            self.before_syntax(variable, Space.Location.VARIABLE_PREFIX, p)
            p.append("as")
            self.visit(variable.name, p)
            self.after_syntax(variable, p)

        self.after_syntax(multi_variable, p)
        self.visit(ca.body, p)
        self.after_syntax(ca, p)
        return ca

    def visit_class_declaration(self, class_decl: ClassDeclaration, p: PrintOutputCapture[P]) -> J:
        self.before_syntax(class_decl, Space.Location.CLASS_DECLARATION_PREFIX, p)
        self.visit_space(Space.EMPTY, Space.Location.ANNOTATIONS, p)
        self.visit(class_decl.leading_annotations, p)
        self.visit(class_decl.padding.kind.annotations, p)
        self.visit_space(class_decl.padding.kind.prefix,
                         Space.Location.CLASS_KIND, p)
        p.append("class")
        self.visit(class_decl.name, p)

        if class_decl.padding.implements is not None:
            omit_parens = class_decl.padding.implements.markers.find_first(
                OmitParentheses) is not None
            self._print_container("" if omit_parens else "(",
                                  class_decl.padding.implements,
                                  JContainer.Location.IMPLEMENTS, ",",
                                          "" if omit_parens else ")", p)

        self.visit(class_decl.body, p)
        self.after_syntax(class_decl, p)
        return class_decl

    def visit_control_parentheses(self, control_parens: ControlParentheses[J], p: PrintOutputCapture[P]) -> J:
        self.before_syntax(control_parens, Space.Location.CONTROL_PARENTHESES_PREFIX, p)
        self._print_right_padded(control_parens.padding.tree,
                                 JRightPadded.Location.PARENTHESES, "", p)
        self.after_syntax(control_parens, p)
        return control_parens

    def visit_else(self, else_: If.Else, p: PrintOutputCapture[P]) -> J:
        self.before_syntax(else_, Space.Location.ELSE_PREFIX, p)
        parent = self.cursor.parent_tree_cursor().value
        if isinstance(parent, If) and isinstance(else_.body, If):
            p.append("el")
            self.visit(else_.body, p)
        elif isinstance(else_.body, Block):
            p.append("else")
            self.visit(else_.body, p)
        else:
            p.append("else")
            p.append(':')
            self.visit(else_.body, p)
        self.after_syntax(else_, p)
        return else_

    def visit_for_each_control(self, control: ForEachLoop.Control, p: PrintOutputCapture[P]) -> J:
        self.before_syntax(control, Space.Location.FOR_EACH_CONTROL_PREFIX, p)
        self.visit_right_padded(control.padding.variable,
                                JRightPadded.Location.FOREACH_VARIABLE, p)
        p.append("in")
        self.visit_right_padded(control.padding.iterable,
                                JRightPadded.Location.FOREACH_ITERABLE, p)
        self.after_syntax(control, p)
        return control

    def visit_for_each_loop(self, for_each_loop: ForEachLoop, p: PrintOutputCapture[P]) -> J:
        self.before_syntax(for_each_loop, Space.Location.FOR_EACH_LOOP_PREFIX, p)
        p.append("for")
        self.visit(for_each_loop.control, p)
        self.visit(for_each_loop.body, p)
        self.after_syntax(for_each_loop, p)
        return for_each_loop

    def visit_identifier(self, ident: Identifier, p: PrintOutputCapture[P]) -> J:
        self.before_syntax(ident, Space.Location.IDENTIFIER_PREFIX, p)
        quoted = ident.markers.find_first(Quoted)
        style = {
            Quoted.Style.SINGLE: "'",
            Quoted.Style.DOUBLE: '"',
            Quoted.Style.TRIPLE_SINGLE: "'''",
            Quoted.Style.TRIPLE_DOUBLE: '"""',
        }

        if quoted is not None:
            p.append(style[quoted.style])
        p.append(ident.simple_name)
        if quoted is not None:
            p.append(style[quoted.style])
        self.after_syntax(ident, p)
        return ident

    def visit_if(self, iff: If, p: PrintOutputCapture[P]) -> J:
        self.before_syntax(iff, Space.Location.IF_PREFIX, p)
        p.append("if")
        self.visit(iff.if_condition, p)

        then_part = iff.padding.then_part
        if not isinstance(then_part.element, Block):
            p.append(":")
        self._print_statement(then_part, JRightPadded.Location.IF_THEN, p)
        self.visit(iff.else_part, p)
        self.after_syntax(iff, p)
        return iff

    def visit_import(self, im: Import, p: PrintOutputCapture[P]) -> J:
        self.before_syntax(im, Space.Location.IMPORT_PREFIX, p)
        if isinstance(im.qualid.target, Empty):
            self.visit(im.qualid.name, p)
        else:
            self.visit(im.qualid, p)
        self._print_left_padded("as", im.padding.alias,
                                JLeftPadded.Location.IMPORT_ALIAS_PREFIX, p)
        self.after_syntax(im, p)
        return im

    def visit_lambda(self, lambda_: Lambda, p: PrintOutputCapture[P]) -> J:
        self.before_syntax(lambda_, Space.Location.LAMBDA_PREFIX, p)
        p.append("lambda")
        self.visit_space(lambda_.parameters.prefix,
                         Space.Location.LAMBDA_PARAMETERS_PREFIX, p)
        self.visit_markers(lambda_.parameters.markers, p)
        self._print_right_padded_list(lambda_.parameters.padding.parameters,
                                      JRightPadded.Location.LAMBDA_PARAM, ",", p)
        self.visit_space(lambda_.arrow, Space.Location.LAMBDA_ARROW_PREFIX, p)
        p.append(":")
        self.visit(lambda_.body, p)
        self.after_syntax(lambda_, p)
        return lambda_

    def visit_literal(self, literal: Literal, p: PrintOutputCapture[P]) -> J:
        if literal.value is None and literal.value_source is None:
            # currently, also `...` is mapped to a `None` value
            literal = literal.with_value_source("None")

        self.before_syntax(literal, Space.Location.LITERAL_PREFIX, p)
        unicode_escapes = literal.unicode_escapes
        if unicode_escapes is None:
            p.append(literal.value_source)
        elif literal.value_source is not None:
            surrogate_iter = iter(unicode_escapes)
            surrogate = next(surrogate_iter, None)
            i = 0
            if surrogate is not None and surrogate.value_source_index == 0:
                p.append("\\u").append(surrogate.code_point)
                surrogate = next(surrogate_iter, None)

            value_source_arr = list(literal.value_source)
            for c in value_source_arr:
                p.append(c)
                if surrogate is not None and surrogate.value_source_index == i + 1:
                    while surrogate is not None and surrogate.value_source_index == i + 1:
                        p.append("\\u").append(surrogate.code_point)
                        surrogate = next(surrogate_iter, None)
                    i += 1

        self.after_syntax(literal, p)
        return literal

    def visit_marker(self, marker: Marker, p: PrintOutputCapture[P]) -> Marker:
        return self.delegate.visit_marker(marker, p)

    def visit_method_declaration(self, method: MethodDeclaration, p: PrintOutputCapture[P]) -> J:
        self.before_syntax(method, Space.Location.METHOD_DECLARATION_PREFIX, p)
        self.visit_space(Space.EMPTY, Space.Location.ANNOTATIONS, p)
        self.visit(method.leading_annotations, p)
        for m in method.modifiers:
            self.visit_modifier(m, p)

        # TODO: In java the method name is a J.Literal, which we can visit here its a string.
        # Maybe we won't need to visit the name at all.
        self.visit(method._name.identifier, p)
        self._print_container("(", method.padding.parameters,
                              JContainer.Location.METHOD_DECLARATION_PARAMETERS, ",", ")", p)
        self.visit(method.return_type_expression, p)
        self.visit(method.body, p)
        self.after_syntax(method, p)
        return method

    def visit_method_invocation(self, method: MethodInvocation, p: PrintOutputCapture[P]) -> J:
        self.before_syntax(method, Space.Location.METHOD_INVOCATION_PREFIX, p)
        self._print_right_padded(method.padding.select,
                                 JRightPadded.Location.METHOD_SELECT,
                                         "" if method.name.simple_name == "" else ".", p)
        self._print_container("<", method.padding.type_parameters,
                              JContainer.Location.TYPE_PARAMETERS, ",", ">", p)
        self.visit(method.name, p)

        if method.markers.find_first(OmitParentheses) is not None:
            before = ""
            after = ""
        else:
            before = "("
            after = ")"

        self._print_container(before, method.padding.arguments,
                              JContainer.Location.METHOD_INVOCATION_ARGUMENTS, ",", after, p)
        self.after_syntax(method, p)
        return method

    def visit_modifier(self, mod: Modifier, p: PrintOutputCapture[P]) -> J:
        keyword_map = {
            Modifier.Type.Default: 'def',
            Modifier.Type.Async: 'async'
        }

        keyword = keyword_map.get(mod.type, None)
        if not keyword:
            raise ValueError(f"Unknown modifier: {mod.type}")

        if keyword:
            self.visit(mod.annotations, p)
            self.before_syntax(mod, Space.Location.MODIFIER_PREFIX, p)
            p.append(keyword)
            self.after_syntax(mod, p)
        return mod

    def visit_new_array(self, new_array: NewArray, p: PrintOutputCapture[P]) -> J:
        self.before_syntax(new_array, Space.Location.NEW_ARRAY_PREFIX, p)
        self._print_container("[", new_array.padding.initializer,
                              JContainer.Location.NEW_ARRAY_INITIALIZER, ",", "]", p)
        self.after_syntax(new_array, p)
        return new_array

    def visit_parameterized_type(self, type_: ParameterizedType, p: PrintOutputCapture[P]) -> J:
        self.before_syntax(type_, Space.Location.PARAMETERIZED_TYPE_PREFIX, p)
        self.visit(type_.clazz, p)
        self._print_container("[", type_.padding.type_parameters,
                              JContainer.Location.TYPE_PARAMETERS, ",", "]", p)
        self.after_syntax(type_, p)
        return type_

    def visit_switch(self, sw: Switch, p: PrintOutputCapture[P]) -> J:
        self.before_syntax(sw, Space.Location.SWITCH_PREFIX, p)
        p.append("match")
        self.visit(sw.selector, p)
        self.visit(sw.cases, p)
        self.after_syntax(sw, p)
        return sw

    def visit_ternary(self, ternary: Ternary, p: PrintOutputCapture[P]) -> J:
        self.before_syntax(ternary, Space.Location.TERNARY_PREFIX, p)
        self.visit(ternary.true_part, p)
        self.visit_space(ternary.padding.true_part.before,
                         Space.Location.TERNARY_TRUE, p)
        p.append("if")
        self.visit(ternary.condition, p)
        self._print_left_padded("else", ternary.padding.false_part,
                                JLeftPadded.Location.TERNARY_FALSE, p)
        self.after_syntax(ternary, p)
        return ternary

    def visit_throw(self, thrown: Throw, p: PrintOutputCapture[P]) -> J:
        self.before_syntax(thrown, Space.Location.THROW_PREFIX, p)
        p.append("raise")
        self.visit(thrown.exception, p)
        self.after_syntax(thrown, p)
        return thrown

    def visit_try(self, tryable: Try, p: PrintOutputCapture[P]) -> J:
        is_with_statement = tryable.resources is not None and len(tryable.resources) > 0

        self.before_syntax(tryable, Space.Location.TRY_PREFIX, p)
        p.append("with") if is_with_statement else p.append("try")

        resources = tryable.padding.resources
        if is_with_statement and resources is not None:
            self.visit_space(resources.before, Space.Location.TRY_RESOURCES, p)
            omit_parentheses = resources.markers.find_first(OmitParentheses) is not None

            if not omit_parentheses:
                p.append("(")

            first = True
            for resource in resources.padding.elements:
                if not first:
                    p.append(",")
                else:
                    first = False

                self.visit_space(resource.element.prefix, Space.Location.TRY_RESOURCE, p)
                self.visit_markers(resource.element.markers, p)

                decl = resource.element.variable_declarations
                if isinstance(decl, Assignment):
                    assignment = decl
                    self.visit(assignment.assignment, p)
                    if not isinstance(assignment.variable, Empty):
                        self.visit_space(assignment.padding.assignment.before,
                                         Space.Location.LANGUAGE_EXTENSION, p)
                        p.append("as")
                        self.visit(assignment.variable, p)
                else:
                    self.visit(decl, p)

                self.visit_space(resource.after, Space.Location.TRY_RESOURCE_SUFFIX, p)
                self.visit_markers(resource.markers, p)

            self.visit_markers(resources.markers, p)
            if not omit_parentheses:
                p.append(")")

        try_body = tryable.body
        parent_cursor = self.cursor.parent_tree_cursor()
        parent_value = parent_cursor.value
        else_wrapper = parent_value if isinstance(parent_value, TrailingElseWrapper) else None

        self.visit(try_body, p)
        self.visit(tryable.catches, p)

        if else_wrapper is not None:
            self.visit_space(else_wrapper.padding.else_block.before,
                             Space.Location.ELSE_PREFIX, p)
            p.append("else")
            self.visit(else_wrapper.else_block, p)

        self._print_left_padded("finally", tryable.padding.finally_, JLeftPadded.Location.TRY_FINALLY, p)
        self.after_syntax(tryable, p)
        return tryable

    def visit_unary(self, unary: Unary, p: PrintOutputCapture[P]) -> J:
        self.before_syntax(unary, Space.Location.UNARY_PREFIX, p)
        operator_map = {
            Unary.Type.Not: 'not',
            Unary.Type.Positive: '+',
            Unary.Type.Negative: '-',
            Unary.Type.Complement: '~'
        }
        op = operator_map.get(unary.operator, '')
        if not op:
            raise ValueError(f"Unknown unary operator: {unary.operator}")

        p.append(op)
        self.visit(unary.expression, p)
        self.after_syntax(unary, p)
        return unary

    def visit_variable(self, variable: VariableDeclarations.NamedVariable, p: PrintOutputCapture[P]) -> J:
        self.before_syntax(variable, Space.Location.VARIABLE_PREFIX, p)
        vd = cast(VariableDeclarations, self.cursor.parent_tree_cursor().value)
        padding = cast(JRightPadded[VariableDeclarations.NamedVariable],
                       self.cursor.parent.value)  # pyright: ignore [reportOptionalMemberAccess]
        type_ = vd.type_expression

        if isinstance(type_, SpecialParameter):
            special = type_
            self.visit(special, p)
            type_ = special.type_hint

        if variable.name.simple_name == "":
            self.visit(variable.initializer, p)
        else:
            if vd.varargs is not None:
                self.visit_space(vd.varargs, Space.Location.VARARGS, p)
                p.append('*')
            if vd.markers.find_first(KeywordArguments) is not None:
                p.append("**")
            self.visit(variable.name, p)
            if type_ is not None:
                self.visit_space(padding.after,
                                 JRightPadded.Location.NAMED_VARIABLE.after_location, p)
                p.append(':')
                self.visit(type_, p)
            self._print_left_padded("=", variable.padding.initializer,
                                    JLeftPadded.Location.VARIABLE_INITIALIZER, p)
        self.after_syntax(variable, p)
        return variable

    def visit_variable_declarations(self, multi_variable: VariableDeclarations, p: PrintOutputCapture[P]) -> J:
        self.before_syntax(multi_variable, Space.Location.VARIABLE_DECLARATIONS_PREFIX, p)
        self.visit_space(Space.EMPTY, Space.Location.ANNOTATIONS, p)
        self.visit(multi_variable.leading_annotations, p)

        for m in multi_variable.modifiers:
            self.visit_modifier(m, p)

        if multi_variable.markers.find_first(KeywordOnlyArguments) is not None:
            p.append("*")

        nodes = multi_variable.padding.variables
        for i, node in enumerate(nodes):
            self.cursor = Cursor(self.cursor, node)
            self.visit(node.element, p)
            self.visit_markers(node.markers, p)
            if i < len(nodes) - 1:
                p.append(",")
            self.cursor = self.cursor.parent  # pyright: ignore [reportAttributeAccessIssue]

        self.after_syntax(multi_variable, p)
        return multi_variable



__path__ = __import__('pkgutil').extend_path(__path__, __name__)

from .support_types import *
from .parser import *
from .visitor import *
from .markers import *
from .style import *
from .format import *
from .templating import *

__all__ = [
    'Py',
    'PyContainer',
    'PyLeftPadded',
    'PyRightPadded',
    'PySpace',
    'PythonParser',
    'PythonParserBuilder',
    'PythonVisitor',

    # Markers
    'KeywordArguments',
    'KeywordOnlyArguments',
    'Quoted',

    # AST types
    'Async',
    'Await',
    'Binary',
    'ChainedAssignment',
    'CollectionLiteral',
    'CompilationUnit',
    'ComprehensionExpression',
    'Del',
    'DictLiteral',
    'ErrorFrom',
    'ExceptionType',
    'ExpressionStatement',
    'ExpressionTypeTree',
    'ForLoop',
    'FormattedString',
    'KeyValue',
    'LiteralType',
    'MatchCase',
    'MultiImport',
    'NamedArgument',
    'Pass',
    'PyComment',
    'Slice',
    'SpecialParameter',
    'Star',
    'StatementExpression',
    'TrailingElseWrapper',
    'TypeAlias',
    'TypeHint',
    'TypeHintedExpression',
    'UnionType',
    'VariableScope',
    'YieldFrom',

    # Style
    'PythonStyle',
    'SpacesStyle',
    'TabsAndIndentsStyle',
    'WrappingAndBracesStyle',
    'BlankLinesStyle',
    'OtherStyle',
    'IntelliJ',

    # Formatter
    'AutoFormat',
    'BlankLinesVisitor',
    'MinimumViableSpacingVisitor',
    'NormalizeFormatVisitor',
    'NormalizeTabsOrSpacesVisitor',
    'SpacesVisitor',

    # Templating
    'PythonTemplate',
]



from __future__ import annotations

import sys
import textwrap
from enum import Enum, auto
from typing import TypeVar, Optional, Union, cast, List

from rewrite import Tree, Cursor, list_map, PrintOutputCapture, RecipeRunException
from rewrite.java import J, Space, JRightPadded, JLeftPadded, JContainer, JavaSourceFile, \
    Block, Label, ArrayDimension, ClassDeclaration, Empty, MethodDeclaration, \
    Binary, MethodInvocation, FieldAccess, Identifier, Lambda, Comment, TrailingComma, Expression, NewArray, \
    Annotation, Literal, If
from rewrite.python import PythonVisitor, TabsAndIndentsStyle, PySpace, PyContainer, PyRightPadded, DictLiteral, \
    CollectionLiteral, ExpressionStatement, OtherStyle, IntelliJ, ComprehensionExpression, PyComment
from rewrite.visitor import P, T, TreeVisitor

J2 = TypeVar('J2', bound=J)


class TabsAndIndentsVisitor(PythonVisitor[P]):

    def __init__(self, style: TabsAndIndentsStyle, other: OtherStyle = IntelliJ.other(), stop_after: Optional[Tree] = None):
        self._stop_after = stop_after
        self._style = style
        self._other = other
        self._stop = False

    def visit(self, tree: Optional[Tree], p: P, parent: Optional[Cursor] = None) -> Optional[J]:
        if parent is not None:
            self._cursor = parent
            if tree is None:
                return cast(Optional[J], self.default_value(None, p))

            for c in parent.get_path_as_cursors():
                v = c.value
                space = None
                if isinstance(v, J):
                    space = v.prefix
                elif isinstance(v, JRightPadded):
                    space = v.after
                elif isinstance(v, JLeftPadded):
                    space = v.before
                elif isinstance(v, JContainer):
                    space = v.before

                if space is not None and '\n' in space.last_whitespace:
                    indent = self.find_indent(space)
                    if indent != 0:
                        c.put_message("last_indent", indent)

            for next_parent in parent.get_path():
                if isinstance(next_parent, J):
                    self.pre_visit(next_parent, p)
                    break

        return super().visit(tree, p)

    def pre_visit(self, tree: T, p: P) -> Optional[T]:
        if isinstance(tree, (JavaSourceFile, Label, ArrayDimension, ClassDeclaration)):
            self.cursor.put_message("indent_type", self.IndentType.ALIGN)
        elif isinstance(tree, (Block, If)):
            self.cursor.put_message("indent_type", self.IndentType.INDENT)
        elif isinstance(tree, (DictLiteral, CollectionLiteral, NewArray, ComprehensionExpression)):
            self.cursor.put_message("indent_type", self.IndentType.CONTINUATION_INDENT
            if self._other.use_continuation_indent.collections_and_comprehensions else self.IndentType.INDENT)
        elif isinstance(tree, Expression) and not isinstance(tree, ExpressionStatement):
            self.cursor.put_message("indent_type", self.IndentType.INDENT)

        return tree

    def post_visit(self, tree: T, p: P) -> Optional[T]:
        if self._stop_after and tree == self._stop_after:
            self._stop = True
        return tree

    def visit_space(self, space: Optional[Space], loc: Optional[Union[PySpace.Location, Space.Location]],
                    p: P) -> Space:
        if space is None:
            return space  # type: ignore

        self._cursor.put_message("last_location", loc)
        parent = self._cursor.parent
        align_to_annotation = False

        if parent is not None:
            if isinstance(parent.value, Annotation):
                parent.parent_or_throw.put_message("after_annotation", True)
            elif not any([isinstance(_, Annotation) for _ in parent.get_path()]):
                align_to_annotation = self.cursor.poll_nearest_message("after_annotation", False)

        if loc == Space.Location.METHOD_SELECT_SUFFIX:
            chained_indent = cast(int, self.cursor.parent_tree_cursor().get_message("chained_indent", None))
            if chained_indent is not None:
                self.cursor.parent_tree_cursor().put_message("last_indent", chained_indent)
                return self._indent_to(space, chained_indent, loc)

        indent = cast(int, self.cursor.get_nearest_message("last_indent")) or 0
        indent_type = self.cursor.parent_or_throw.get_nearest_message("indent_type") or self.IndentType.ALIGN

        if not space.comments and '\n' not in space.last_whitespace or parent is None:
            return space

        cursor_value = self._cursor.value

        # Block spaces are always aligned to their parent
        # The second condition ensure init blocks are ignored.
        # TODO: Second condition might be removed since it's not relevant for Python
        align_block_prefix_to_parent = loc is Space.Location.BLOCK_PREFIX and '\n' in space.whitespace and \
                                       (isinstance(cursor_value, Block) and not isinstance(
                                           self.cursor.parent_tree_cursor().value, Block))

        align_block_to_parent = loc in (
            Space.Location.NEW_ARRAY_INITIALIZER_SUFFIX,
            Space.Location.CATCH_PREFIX,
            Space.Location.TRY_FINALLY,
            Space.Location.ELSE_PREFIX,
        )

        if (loc == Space.Location.EXTENDS and "\n" in space.whitespace) or \
                Space.Location.EXTENDS == self.cursor.parent_or_throw.get_message("last_location", None):
            indent_type = self.IndentType.CONTINUATION_INDENT

        if align_block_prefix_to_parent or align_block_to_parent or align_to_annotation:
            indent_type = self.IndentType.ALIGN

        if indent_type == self.IndentType.INDENT:
            indent += self._style.indent_size
        elif indent_type == self.IndentType.CONTINUATION_INDENT:
            indent += self._style.continuation_indent

        s: Space = self._indent_to(space, indent, loc)
        if isinstance(cursor_value, J):
            self.cursor.put_message("last_indent", indent)
        elif loc == Space.Location.METHOD_SELECT_SUFFIX:
            self.cursor.parent_tree_cursor().put_message("last_indent", indent)

        return s

    @staticmethod
    def compute_first_parameter_offset(method: MethodDeclaration, first_arg: J, cursor: Cursor) -> int:
        # Clear any annotations to avoid them affecting the offset calculation
        method = method.with_leading_annotations([])
        class FirstArgPrinter(PrintOutputCapture[TreeVisitor]):
            def append(self, text: Optional[str] = None) -> PrintOutputCapture[P]:
                if self._context.cursor.value is first_arg:
                    raise RecipeRunException()
                return super().append(text)

        printer = method.printer(cursor)
        capture = FirstArgPrinter(printer)
        try:
            printer.visit(method, capture, cursor.parent_or_throw)
        except RecipeRunException:
            source = capture.get_out()
            def_idx = source.index("def")
            async_idx = source.find("async")
            start_idx = async_idx if async_idx != -1 and async_idx < def_idx else def_idx
            return len(source[start_idx:]) + len(first_arg.prefix.last_whitespace)

    def visit_right_padded(self, right: Optional[JRightPadded[T]],
                           loc: Union[PyRightPadded.Location, JRightPadded.Location], p: P) -> Optional[
        JRightPadded[T]]:

        if right is None:
            return None

        self.cursor = Cursor(self._cursor, right)

        indent: int = cast(int, self.cursor.get_nearest_message("last_indent")) or 0

        t: T = right.element
        after = right.after
        # TODO: Check if the visit_and_cast is really required here

        if isinstance(t, J):
            elem = t
            trailing_comma = right.markers.find_first(TrailingComma)
            if '\n' in right.after.last_whitespace or '\n' in elem.prefix.last_whitespace:
                if loc in (JRightPadded.Location.FOR_CONDITION,
                           JRightPadded.Location.FOR_UPDATE):
                    raise ValueError("This case should not be possible, should be safe for removal...")
                elif loc in (JRightPadded.Location.METHOD_DECLARATION_PARAMETER,
                             JRightPadded.Location.RECORD_STATE_VECTOR):
                    if isinstance(elem, Empty):
                        elem = elem.with_prefix(self._indent_to(elem.prefix, indent, loc.after_location))
                        after = right.after
                    else:
                        container: JContainer[J] = cast(JContainer[J], self.cursor.parent_or_throw.value)
                        elements: List[J] = container.elements
                        first_arg: J = elements[0]
                        last_arg: J = elements[-1]

                        # TODO: style.MethodDeclarationParameters doesn't exist for Python
                        # but should be self._style.method_declaration_parameters.align_when_multiple
                        if self._style.method_declaration_parameters.align_multiline_parameters:
                            method = self.cursor.first_enclosing(MethodDeclaration)
                            if method is not None:
                                if "\n" in first_arg.prefix.last_whitespace:
                                    if self._other.use_continuation_indent.method_declaration_parameters:
                                        align_to = indent + self._style.continuation_indent
                                    else:
                                        align_to = self._get_length_of_whitespace(first_arg.prefix.last_whitespace)
                                else:
                                    align_to = indent + self.compute_first_parameter_offset(method, first_arg,
                                                                                            self.cursor)
                                self.cursor.parent_or_throw.put_message("last_indent", align_to - self._style.continuation_indent)
                                elem = self.visit_and_cast(elem, J, p)
                                self.cursor.parent_or_throw.put_message("last_indent", indent)
                                after = self._indent_to(right.after, indent if t is last_arg else align_to, loc.after_location)
                            else:
                                after = right.after
                        else:
                            elem = self.visit_and_cast(elem, J, p)
                            after = self._indent_to(right.after,
                                                    indent if t is last_arg else self._style.continuation_indent,
                                                    loc.after_location)

                elif loc == JRightPadded.Location.METHOD_INVOCATION_ARGUMENT:
                    elem, after = self._visit_method_invocation_argument_j_type(elem, right, indent, loc, p)
                elif loc in (JRightPadded.Location.NEW_CLASS_ARGUMENTS,
                             JRightPadded.Location.ARRAY_INDEX,
                             JRightPadded.Location.PARENTHESES,
                             JRightPadded.Location.TYPE_PARAMETER):
                    elem = self.visit_and_cast(elem, J, p)
                    after = self._indent_to(right.after, indent, loc.after_location)
                elif loc in (PyRightPadded.Location.COLLECTION_LITERAL_ELEMENT, PyRightPadded.Location.DICT_LITERAL_ELEMENT):
                    elem = self.visit_and_cast(elem, J, p)
                    args = cast(JContainer[J], self.cursor.parent_or_throw.value)
                    if not trailing_comma and args.padding.elements[-1] is right:
                        self.cursor.parent_or_throw.put_message("indent_type", self.IndentType.ALIGN)
                    after = self.visit_space(right.after, loc.after_location, p)
                    if trailing_comma:
                        self.cursor.parent_or_throw.put_message("indent_type", self.IndentType.ALIGN)
                        trailing_comma = trailing_comma.with_suffix(self.visit_space(trailing_comma.suffix, loc.after_location, p))
                        right = right.with_markers(right.markers.compute_by_type(TrailingComma, lambda t: trailing_comma))
                elif loc == JRightPadded.Location.ANNOTATION_ARGUMENT:
                    raise NotImplementedError("Annotation argument not implemented")
                else:
                    elem = self.visit_and_cast(elem, J, p)
                    after = self.visit_space(right.after, loc.after_location, p)
            else:
                if loc in (JRightPadded.Location.NEW_CLASS_ARGUMENTS, JRightPadded.Location.METHOD_INVOCATION_ARGUMENT):
                    any_other_arg_on_own_line = False
                    if "\n" not in elem.prefix.last_whitespace:
                        args = cast(JContainer[J], self.cursor.parent_or_throw.value)
                        for arg in args.padding.elements:
                            if arg == self.cursor.value:
                                continue
                            if "\n" in arg.element.prefix.last_whitespace:
                                any_other_arg_on_own_line = True
                                break
                        if not any_other_arg_on_own_line:
                            elem = self.visit_and_cast(elem, J, p)
                            after = self._indent_to(right.after, indent, loc.after_location)

                    if not any_other_arg_on_own_line:
                        if not isinstance(elem, Binary):
                            if not isinstance(elem, MethodInvocation) or "\n" in elem.prefix.last_whitespace:
                                self.cursor.put_message("last_indent", indent + self._style.continuation_indent)
                            else:
                                method_invocation = elem
                                select = method_invocation.select
                                if isinstance(select, (FieldAccess, Identifier, MethodInvocation)):
                                    self.cursor.put_message("last_indent", indent + self._style.continuation_indent)

                        elem = self.visit_and_cast(elem, J, p)
                        after = self.visit_space(right.after, loc.after_location, p)
                else:
                    elem = self.visit_and_cast(elem, J, p)
                    after = self.visit_space(right.after, loc.after_location, p)

            t = cast(T, elem)
        else:
            after = self.visit_space(right.after, loc.after_location, p)

        self.cursor = self.cursor.parent  # type: ignore
        return right.with_after(after).with_element(t)

    def visit_container(self, container: Optional[JContainer[J2]],
                        loc: Union[PyContainer.Location, JContainer.Location], p: P) -> JContainer[J2]:
        if container is None:
            return container  # type: ignore

        self._cursor = Cursor(self._cursor, container)

        indent = cast(int, self.cursor.get_nearest_message("last_indent")) or 0
        if '\n' in container.before.last_whitespace:
            if loc in (JContainer.Location.TYPE_PARAMETERS,
                       JContainer.Location.IMPLEMENTS,
                       JContainer.Location.THROWS,
                       JContainer.Location.NEW_CLASS_ARGUMENTS,
                       JContainer.Location.METHOD_DECLARATION_PARAMETERS):
                before = self._indent_to(container.before, indent + self._style.continuation_indent,
                                         loc.before_location)
                self.cursor.put_message("indent_type", self.IndentType.ALIGN)
                self.cursor.put_message("last_indent", indent + self._style.continuation_indent)
            else:
                before = self.visit_space(container.before, loc.before_location, p)
            js = list_map(lambda t: self.visit_right_padded(t, loc.element_location, p), container.padding.elements)
        else:
            if loc == JContainer.Location.METHOD_DECLARATION_PARAMETERS:
                self.cursor.put_message("indent_type", self.IndentType.CONTINUATION_INDENT if self._other.use_continuation_indent.method_declaration_parameters else self.IndentType.INDENT)
            elif loc == JContainer.Location.METHOD_INVOCATION_ARGUMENTS:
                self.cursor.put_message("indent_type", self.IndentType.CONTINUATION_INDENT if self._other.use_continuation_indent.method_call_arguments else self.IndentType.INDENT)
            before = self.visit_space(container.before, loc.before_location, p)
            js = list_map(lambda t: self.visit_right_padded(t, loc.element_location, p), container.padding.elements)

        self._cursor = self._cursor.parent  # type: ignore

        if container.padding.elements is js and container.before is before:
            return container
        return JContainer(before, js, container.markers)

    @staticmethod
    def _is_doc_comment(expression_statement: ExpressionStatement, cursor: Cursor) -> bool:
        expr = expression_statement.expression
        return isinstance(expr, Literal) and isinstance(expr.value_source, str) and (
                (expr.value_source.startswith('"""') and expr.value_source.endswith('"""')) or
                (expr.value_source.startswith("'''") and expr.value_source.endswith("'''"))) and \
            cursor.first_enclosing(Block) is not None

    def visit_expression_statement(self, expression_statement: ExpressionStatement, p: P) -> J:
        if self._is_doc_comment(expression_statement, self.cursor):
            prefix_before = len(expression_statement.prefix.last_whitespace.split("\n")[-1])
            stm = cast(ExpressionStatement, super().visit_expression_statement(expression_statement, p))
            literal = cast(Literal, stm.expression)
            shift = len(stm.prefix.last_whitespace.split("\n")[-1]) - prefix_before
            return stm.with_expression(
                literal.with_value_source(textwrap.indent(str(literal.value_source), shift * " ")[shift:]))

        return super().visit_expression_statement(expression_statement, p)

    def _indent_to(self, space: Space, column: int, space_location: Optional[Union[PySpace.Location, Space.Location]]) -> Space:
        s = space
        whitespace = s.whitespace

        if space_location == Space.Location.COMPILATION_UNIT_PREFIX and whitespace:
            s = s.with_whitespace("")
        elif not s.comments and "\n" not in s.last_whitespace:
            return s

        if not s.comments:
            indent = self.find_indent(s)
            if indent != column:
                shift = column - indent
                s = s.with_whitespace(self._indent(whitespace, shift))
        else:
            def whitespace_indent(text: str) -> str:
                # TODO: Placeholder function, taken from java openrewrite.StringUtils
                indent: List[str] = []
                for c in text:
                    if c == '\n' or c == '\r':
                        return ''.join(indent)
                    elif c.isspace():
                        indent.append(c)
                    else:
                        return ''.join(indent)
                return ''.join(indent)

            # TODO: This is the java version, however the python version is probably different
            has_file_leading_comment = space.comments and (
                    (space_location == Space.Location.COMPILATION_UNIT_PREFIX) or (
                        space_location == Space.Location.BLOCK_END) or
                    (space_location == Space.Location.CLASS_DECLARATION_PREFIX and space.comments[0].multiline)
            )

            final_column = column + self._style.indent_size if space_location == Space.Location.BLOCK_END else column
            last_indent: str = space.whitespace[space.whitespace.rfind('\n') + 1:]
            indent = self._get_length_of_whitespace(whitespace_indent(last_indent))

            if indent != final_column or s.comments:
                if (has_file_leading_comment or ("\n" in whitespace)) and (
                        # Do not shift single-line comments at column 0.
                        not (s.comments and isinstance(s.comments[0], PyComment) and
                             not s.comments[0].multiline and self._get_length_of_whitespace(s.whitespace) == 0)):
                    shift = final_column - indent
                    s = s.with_whitespace(whitespace[:whitespace.rfind('\n') + 1] + self._indent(last_indent, shift))

                final_space = s
                last_comment_pos = len(s.comments) - 1

                def _process_comment(i: int, c: Comment) -> Comment:
                    if isinstance(c, PyComment) and not c.multiline:
                        # Do not shift single line comments at col 0.
                        if i != last_comment_pos and self._get_length_of_whitespace(c.suffix) == 0:
                            return c

                    prior_suffix = space.whitespace if i == 0 else final_space.comments[i - 1].suffix

                    if space_location == Space.Location.BLOCK_END and i != len(final_space.comments) - 1:
                        to_column = column + self._style.indent_size
                    else:
                        to_column = column

                    new_c = c
                    if "\n" in prior_suffix or has_file_leading_comment:
                        new_c = c

                    if '\n' in new_c.suffix:
                        suffix_indent = self._get_length_of_whitespace(new_c.suffix)
                        shift = to_column - suffix_indent
                        new_c = new_c.with_suffix(self._indent(new_c.suffix, shift))

                    return new_c

                s = s.with_comments(list_map(lambda i, c: _process_comment(c, i), s.comments))
        return s

    def _indent(self, whitespace: str, shift: int) -> str:
        return self._shift(whitespace, shift)

    def _shift(self, text: str, shift: int) -> str:
        tab_indent = self._style.tab_size
        if not self._style.use_tab_character:
            tab_indent = sys.maxsize

        if shift > 0:
            text += '\t' * (shift // tab_indent)
            text += ' ' * (shift % tab_indent)
        else:
            if self._style.use_tab_character:
                len_text = len(text) + (shift // tab_indent)
            else:
                len_text = len(text) + shift
            if len_text >= 0:
                text = text[:len_text]

        return text

    def find_indent(self, space: Space) -> int:
        return self._get_length_of_whitespace(space.indent)

    def _get_length_of_whitespace(self, whitespace: Optional[str]) -> int:
        if whitespace is None:
            return 0
        length = 0
        for c in whitespace:
            length += self._style.tab_size if c == '\t' else 1
            if c in ('\n', '\r'):
                length = 0
        return length

    def _visit_method_invocation_argument_j_type(self, elem: J, right: JRightPadded[T], indent: int, loc: Union[PyRightPadded.Location, JRightPadded.Location], p: P) -> tuple[J, Space]:
        if "\n" not in elem.prefix.last_whitespace and isinstance(elem, Lambda):
            body = elem.body
            if not isinstance(body, Binary):
                if "\n" not in body.prefix.last_whitespace:
                    self.cursor.parent_or_throw.put_message("last_indent", indent + self._style.continuation_indent)

        elem = self.visit_and_cast(elem, J, p)
        after = self._indent_to(right.after, indent, loc.after_location)
        if after.comments or "\n" in after.last_whitespace:
            parent = self.cursor.parent_tree_cursor()
            grandparent = parent.parent_tree_cursor()
            # propagate indentation up in the method chain hierarchy
            if isinstance(grandparent.value, MethodInvocation) and grandparent.value.select == parent.value:
                grandparent.put_message("last_indent", indent)
                grandparent.put_message("chained_indent", indent)
        return elem, after

    class IndentType(Enum):
        ALIGN = auto()
        INDENT = auto()
        CONTINUATION_INDENT = auto()



from typing import Optional

from rewrite import Cursor
from rewrite import Tree
from rewrite.java import Space, Literal, Identifier, JRightPadded, JLeftPadded, Modifier
from rewrite.python import PythonVisitor
from rewrite.visitor import T, P


class TreeVisitingPrinter(PythonVisitor):
    INDENT = "  "
    ELEMENT_PREFIX = "\\---"
    CONTINUE_PREFIX = "|---"
    UNVISITED_PREFIX = "#"
    BRANCH_CONTINUE_CHAR = '|'
    BRANCH_END_CHAR = '\\'
    CONTENT_MAX_LENGTH = 120

    _last_cursor_stack = []
    _lines = []

    def __init__(self, indent: str = "  "):
        super().__init__()
        self.INDENT = indent

    def visit(self, tree: Optional[Tree], p: P, parent: Optional[Cursor] = None) -> Optional[T]:
        if tree is None:
            return super().visit(None, p, parent)  # pyright: ignore [reportReturnType]

        _current_stack = list(self._cursor.get_path())
        _current_stack.reverse()
        depth = len(_current_stack)
        if not self._last_cursor_stack:
            self._last_cursor_stack = _current_stack + [tree]
        else:
            diff_position = self.find_diff_pos(_current_stack, self._last_cursor_stack)
            if diff_position >= 0:
                for i in _current_stack[diff_position:]:
                    self._lines += [[depth, i]]
                self._last_cursor_stack = self._last_cursor_stack[:diff_position]

        self._lines += [[depth, tree]]
        self._last_cursor_stack = _current_stack + [tree]
        return super().visit(tree, p, parent)  # pyright: ignore [reportReturnType]

    def _print_tree(self) -> str:
        output = ""
        offset = 0
        for idx, (depth, element) in enumerate(self._lines):
            offset = depth if idx == 0 else offset
            padding = self.INDENT * (depth - offset)
            if idx + 1 < len(self._lines) and self._lines[idx + 1][0] <= depth or idx + 1 == len(self._lines):
                output += padding + self.CONTINUE_PREFIX + self._print_element(element) + "\n"
            else:
                output += padding + self.ELEMENT_PREFIX + self._print_element(element) + "\n"
        return output

    def _print_element(self, element) -> str:
        type_name = type(element).__name__
        line = []

        if hasattr(element, "before"):
            line.append(f"before= {self._print_space(element.before)}")

        if hasattr(element, "after"):
            line.append(f"after= {self._print_space(element.after)}")

        if hasattr(element, "suffix"):
            line.append(f"suffix= {self._print_space(element.suffix)}")

        if hasattr(element, "prefix"):
            line.append(f"prefix= {self._print_space(element.prefix)}")

        if isinstance(element, Identifier):
            type_name = f'{type_name} | "{element.simple_name}"'

        if isinstance(element, Literal):
            type_name = f'{type_name} | {element.value_source}'

        if isinstance(element, JRightPadded):
            return f'{type_name} | after= {self._print_space(element.after)}'

        if isinstance(element, JLeftPadded):
            return f'{type_name} | before= {self._print_space(element.before)}'

        if isinstance(element, Modifier):
            return type_name + (
                (" | " + element.type.name) if hasattr(element, "type") else "")

        if line:
            return type_name + " | " + " | ".join(line)
        return type_name

    @staticmethod
    def _print_space(space: Space) -> str:
        parts = []
        if space.whitespace:
            parts.append(f'whitespace="{repr(space.whitespace)}"')
        if space.comments:
            parts.append(f'comments="{space.comments}"')
        return " ".join(parts).replace("\n", "\\s\n")

    @staticmethod
    def print_tree_all(tree: "Tree") -> str:
        visitor = TreeVisitingPrinter()
        visitor.visit(tree, None, None)
        return visitor._print_tree()

    def find_diff_pos(self, cursor_stack, last_cursor_stack):
        diff_pos = -1
        for i in range(len(cursor_stack)):
            if i >= len(last_cursor_stack):
                diff_pos = i
                break
            if cursor_stack[i] != last_cursor_stack[i]:
                diff_pos = i
                break
        return diff_pos



from __future__ import annotations

import socket
import textwrap
from dataclasses import dataclass, field
from io import StringIO
from pathlib import Path
from typing import Optional, Callable, Iterable, List, TypeVar, Any, cast
from uuid import UUID

from rewrite import InMemoryExecutionContext, ParserInput, ParserBuilder, random_id, ParseError, ParseExceptionResult, \
    ExecutionContext, Recipe, TreeVisitor, SourceFile, PrintOutputCapture
from rewrite.execution import InMemoryLargeSourceSet
from rewrite.python import CompilationUnit
from rewrite.python.parser import PythonParserBuilder
from rewrite.python.printer import PythonPrinter

S = TypeVar('S', bound=SourceFile)

@dataclass(frozen=True, eq=False)
class SourceSpec:
    _id: UUID

    @property
    def id(self) -> UUID:
        return self._id

    _parser: ParserBuilder

    @property
    def parser(self) -> ParserBuilder:
        return self._parser

    _before: str

    @property
    def before(self) -> str:
        return self._before

    _after: Optional[Callable[[str], str]]

    @property
    def after(self) -> Optional[Callable[[str], str]]:
        return self._after

    _source_path: Optional[Path]

    @property
    def source_path(self) -> Optional[Path]:
        return self._source_path

    _after_recipe: Optional[Callable[[S], None]] = lambda _: None

    @property
    def after_recipe(self) -> Optional[Callable[[S], None]]:
        return self._after_recipe


@dataclass(frozen=True)
class CompositeRecipe(Recipe):
    recipes: Iterable[Recipe]

    def get_recipe_list(self) -> List[Recipe]:
        return list(self.recipes)


@dataclass(frozen=True, eq=False)
class RecipeSpec:
    _recipe: Optional[Recipe] = None

    @property
    def recipe(self) -> Optional[Recipe]:
        return self._recipe

    def with_recipe(self, recipe: Recipe) -> RecipeSpec:
        return self if recipe is self._recipe else RecipeSpec(recipe)

    def with_recipes(self, *recipes: Recipe) -> RecipeSpec:
        return RecipeSpec(CompositeRecipe(recipes))

    _parsers: Iterable[ParserBuilder] = field(default_factory=list)

    @property
    def parsers(self) -> Iterable[ParserBuilder]:
        return self._parsers

    def with_parsers(self, parsers: Iterable[ParserBuilder]) -> RecipeSpec:
        return self if parsers is self._parsers else RecipeSpec(self._recipe, parsers)

def rewrite_run(*source_specs: Iterable[SourceSpec], spec: Optional[RecipeSpec] = None) -> None:
    USE_REMOTE = False

    if USE_REMOTE:
        from rewrite_remote import RemotingContext, RemotePrinterFactory
        from rewrite_remote.server import register_remoting_factories
        remoting_context = RemotingContext()
        register_remoting_factories()
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        # if USE_REMOTE:
        s.connect(('localhost', 65432))
        remoting_context.connect(s)
        RemotePrinterFactory(remoting_context.client).set_current()

    try:
        ctx = InMemoryExecutionContext()
        ctx.put_message(ExecutionContext.REQUIRE_PRINT_EQUALS_INPUT, False)

        configured_parsers = {p.source_file_type: p for p in spec.parsers} if spec else {}
        parsers = {s.parser.source_file_type: configured_parsers.get(s.parser.source_file_type, s.parser) for source_spec in source_specs for s in source_spec}

        spec_by_source_file = {}
        for source in source_specs:
            for source_spec in source:
                parser = parsers[source_spec.parser.source_file_type].build()
                source_path = source_spec.source_path if source_spec.source_path \
                    else parser.source_path_from_source_text(Path('.'), source_spec.before)
                for source_file in parser.parse_inputs(
                        [ParserInput(source_path, None, True, lambda: StringIO(source_spec.before))], None, ctx):
                    if isinstance(source_file, ParseError):
                        assert False, f'Parser threw an exception:\n%{source_file.markers.find_first(ParseExceptionResult).message}'  # type: ignore

                    if USE_REMOTE:
                        remoting_context.reset()
                        remoting_context.client.reset()

                    assert source_file.print_all() == source_spec.before

                    spec_by_source_file[source_file] = source_spec

        if spec and spec.recipe:
            recipe = spec.recipe
            before = InMemoryLargeSourceSet(list(spec_by_source_file.keys()))
            result = recipe.run(before, ctx)
            for res in result:
                if res._before and res._after:
                    source_spec = spec_by_source_file[res._before]
                    source_spec.after_recipe(res._after)
                    after_printed = res._after.print_all()
                    if source_spec.after is not None:
                        after = source_spec.after(after_printed)
                        assert after_printed == after
                    else:
                        assert after_printed == source_spec.before
        else:
            for before_source in spec_by_source_file.keys():
                source_spec = spec_by_source_file[before_source]
                if source_spec.after_recipe is not None:
                    source_spec.after_recipe(before_source)

    except Exception:
        raise
    finally:
        if USE_REMOTE:
            remoting_context.close()


def python(before: str, after: Optional[str] = None, after_recipe: Optional[Callable[[CompilationUnit], None]] = lambda s: None) -> list[SourceSpec]:
    return [SourceSpec(
        random_id(),
        PythonParserBuilder(),
        textwrap.dedent(before),
        None if after is None else lambda _: textwrap.dedent(after),
        None,
        cast(Optional[Callable[[S], None]], after_recipe)
    )]


def from_visitor(visitor: TreeVisitor[Any, Any]) -> Recipe:
    return AdHocRecipe(visitor)


@dataclass(frozen=True)
class AdHocRecipe(Recipe):
    visitor: TreeVisitor[Any, Any]

    def get_visitor(self) -> TreeVisitor[Any, Any]:
        return self.visitor



__path__ = __import__('pkgutil').extend_path(__path__, __name__)

from .auto_format import *
from .blank_lines import *
from .normalize_format import *

__all__ = [
    'AutoFormat',
    'BlankLinesVisitor',
    'MinimumViableSpacingVisitor',
    'NormalizeFormatVisitor',
    'NormalizeTabsOrSpacesVisitor',
    'SpacesVisitor',
]
